"""Time handling module.

.. versionadded:: 7.5
"""
#
# (C) Pywikibot team, 2007-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import datetime
import math
import re
import time as _time
import types
from contextlib import suppress

import pywikibot
from pywikibot.tools import PYTHON_VERSION, classproperty, deprecated


__all__ = (
    'parse_duration',
    'str2timedelta',
    'MW_KEYS',
    'Timestamp',
    'TZoneFixedOffset'
)

#: .. versionadded:: 7.5
MW_KEYS = types.MappingProxyType({
    's': 'seconds',
    'h': 'hours',
    'd': 'days',
    'w': 'weeks',
    'y': 'years',
    # 'months' and 'minutes' were removed because confusion outweighs merit
})


class Timestamp(datetime.datetime):

    """Class for handling MediaWiki timestamps.

    This inherits from :python:`datetime.datetime
    <library/datetime.html#datetime.datetime>`, so it can use all of
    the methods and operations of a ``datetime`` object. To ensure that
    the results of any operation are also a Timestamp object, be sure to
    use only Timestamp objects (and :python:`datetime.timedelta
    <library/datetime.html#datetime.timedelta>`) in any operation.

    Use :meth:`Timestamp.fromISOformat` and
    :meth:`Timestamp.fromtimestampformat` to create Timestamp objects
    from MediaWiki string formats. As these constructors are typically
    used to create objects using data passed provided by site and page
    methods, some of which return a Timestamp when previously they
    returned a MediaWiki string representation, these methods also
    accept a Timestamp object, in which case they return a clone.

    Alternatively, :meth:`Timestamp.set_timestamp` can create Timestamp
    objects from :class:`Timestamp`, ``datetime.datetime`` object, or
    strings compliant with ISO8601, MediaWiki, or POSIX formats.

    Use :meth:`Site.server_time()
    <pywikibot.site._apisite.APISite.server_time>` for the current time;
    this is more reliable than using :meth:`Timestamp.utcnow` or
    :meth:`Timestamp.nowutc`.

    .. versionchanged:: 7.5
       moved to :mod:`time` module
    """

    mediawikiTSFormat = '%Y%m%d%H%M%S'  # noqa: N815
    _ISO8601Format_new = '{0:+05d}-{1:02d}-{2:02d}T{3:02d}:{4:02d}:{5:02d}Z'

    @classmethod
    def set_timestamp(
        cls,
        ts: str | datetime.datetime | Timestamp,
    ) -> Timestamp:
        """Set Timestamp from input object.

        ts is converted to a datetime naive object representing UTC time.
        String shall be compliant with:

        - Mediwiki timestamp format: ``YYYYMMDDHHMMSS``
        - ISO8601 format: ``YYYY-MM-DD[T ]HH:MM:SS[Z|±HH[MM[SS[.ffffff]]]]``
        - POSIX format: seconds from Unix epoch ``S{1,13}[.ffffff]]``

        .. versionadded:: 7.5
        .. versionchanged:: 8.0
           raises *TypeError* instead of *ValueError*.

        :param ts: Timestamp, datetime.datetime or str
        :return: Timestamp object
        :raises TypeError: conversion failed
        """
        if isinstance(ts, cls):
            return ts
        if isinstance(ts, datetime.datetime):
            return cls._from_datetime(ts)
        if isinstance(ts, str):
            return cls._from_string(ts)
        raise TypeError(
            f'Unsupported "ts" type, got "{ts}" ({type(ts).__name__})')

    @staticmethod
    def _from_datetime(dt: datetime.datetime) -> Timestamp:
        """Convert a datetime.datetime timestamp to a Timestamp object.

        .. versionadded:: 7.5
        """
        return Timestamp(dt.year, dt.month, dt.day, dt.hour,
                         dt.minute, dt.second, dt.microsecond,
                         dt.tzinfo)

    @classmethod
    def _from_mw(cls, timestr: str) -> Timestamp:
        """Convert a string in MW format to a Timestamp object.

        Mediwiki timestamp format: YYYYMMDDHHMMSS

        .. versionadded:: 7.5
        """
        RE_MW = r'\d{14}'  # noqa: N806
        m = re.fullmatch(RE_MW, timestr)

        if not m:
            raise ValueError(
                f'time data {timestr!r} does not match MW format.')

        return cls.strptime(timestr, cls.mediawikiTSFormat)

    @classmethod
    def _from_iso8601(cls, timestr: str) -> Timestamp:
        """Convert a string in ISO8601 format to a Timestamp object.

        ISO8601 format:
        ``YYYY-MM-DD[T ]HH:MM:SS[[.,]ffffff][Z|±HH[MM[SS[.ffffff]]]]``

        .. versionadded:: 7.5
        """
        RE_ISO8601 = (r'(?:\d{4}-\d{2}-\d{2})(?P<sep>[T ])'  # noqa: N806
                      r'(?:\d{2}:\d{2}:\d{2})(?P<u>[.,]\d{1,6})?'
                      r'(?P<tz>Z|[+\-]\d{2}:?\d{,2})?'
                      )
        m = re.fullmatch(RE_ISO8601, timestr)

        if not m:
            raise ValueError(
                f'time data {timestr!r} does not match ISO8601 format.')

        strpfmt = f'%Y-%m-%d{m["sep"]}%H:%M:%S'
        strpstr = timestr[:19]

        if m['u']:
            strpfmt += '.%f'
            strpstr += m['u'].replace(',', '.')  # .ljust(7, '0')

        if m['tz']:
            if m['tz'] == 'Z':
                strpfmt += 'Z'
                strpstr += 'Z'
            else:
                strpfmt += '%z'
                # strptime wants HHMM, without ':'
                strpstr += (m['tz'].replace(':', '')).ljust(5, '0')

        ts = cls.strptime(strpstr, strpfmt)
        if ts.tzinfo is not None:
            ts = ts.astimezone(datetime.timezone.utc).replace(tzinfo=None)
            # TODO: why pytest in py37 fails without this?
            ts = cls._from_datetime(ts)

        return ts

    @classmethod
    def _from_posix(cls, timestr: str) -> Timestamp:
        """Convert a string in POSIX format to a Timestamp object.

        POSIX format: ``SECONDS[.ffffff]]``

        .. versionadded:: 7.5
        """
        RE_POSIX = r'(?P<S>-?\d{1,13})(?:\.(?P<u>\d{1,6}))?'  # noqa: N806
        m = re.fullmatch(RE_POSIX, timestr)

        if not m:
            msg = "time data '{timestr}' does not match POSIX format."
            raise ValueError(msg.format(timestr=timestr))

        sec = int(m['S'])
        usec = m['u']
        usec = int(usec.ljust(6, '0')) if usec else 0
        if sec < 0 < usec:
            sec -= 1
            usec = 1_000_000 - usec

        return cls(1970, 1, 1) + datetime.timedelta(seconds=sec,
                                                    microseconds=usec)

    @classmethod
    def _from_string(cls, timestr: str) -> Timestamp:
        """Convert a string to a Timestamp object.

        .. versionadded:: 7.5
        """
        handlers = [
            cls._from_mw,
            cls._from_iso8601,
            cls._from_posix,
        ]

        for handler in handlers:
            with suppress(ValueError):
                return handler(timestr)

        raise ValueError(f'time data {timestr!r} does not match any format.')

    @deprecated('replace method', since='8.0.0')  # type: ignore[misc]
    def clone(self) -> Timestamp:
        """Clone this instance.

        .. deprecated:: 8.0
           Use :meth:`replace` method instead.
        """
        return self.replace()

    @classproperty
    def ISO8601Format(cls) -> str:  # noqa: N802
        """ISO8601 format string class property for compatibility purpose."""
        return cls._ISO8601Format()

    @classmethod
    def _ISO8601Format(cls, sep: str = 'T') -> str:  # noqa: N802
        """ISO8601 format string.

        :param sep: one-character separator, placed between the date and
            time
        :return: ISO8601 format string
        """
        assert len(sep) == 1
        return f'%Y-%m-%d{sep}%H:%M:%SZ'

    @classmethod
    def fromISOformat(cls,  # noqa: N802
                      ts: str | Timestamp,
                      sep: str = 'T') -> Timestamp:
        """Convert an ISO 8601 timestamp to a Timestamp object.

        :param ts: ISO 8601 timestamp or a Timestamp object already
        :param sep: one-character separator, placed between the date and
            time
        :return: Timestamp object
        """
        # If inadvertently passed a Timestamp object, use replace()
        # to create a clone.
        if isinstance(ts, cls):
            return ts.replace()

        if not isinstance(ts, str):
            raise TypeError(
                f'ts argument must be a string or a Timestamp object,'
                f' not {type(ts).__name__}')

        return cls._from_iso8601(f'{ts[:10]}{sep}{ts[11:]}')

    @classmethod
    def fromtimestampformat(cls,
                            ts: str | Timestamp,
                            strict: bool = False) -> Timestamp:
        """Convert a MediaWiki internal timestamp to a Timestamp object.

        .. versionchanged:: 3.0
           create a Timestamp if only year, month and day are given.
        .. versionchanged:: 8.0
           the *strict* parameter was added which discards missing
           element tolerance.

        Example
        -------

        >>> Timestamp.fromtimestampformat('20220705082234')
        Timestamp(2022, 7, 5, 8, 22, 34)
        >>> Timestamp.fromtimestampformat('20220927')
        Timestamp(2022, 9, 27, 0, 0)
        >>> Timestamp.fromtimestampformat('20221109', strict=True)
        Traceback (most recent call last):
        ...
        ValueError: time data '20221109' does not match MW format.

        :param ts: the timestamp to be converted
        :param strict: If true, do not ignore missing timestamp elements
            for hours, minutes or seconds
        :return: return the *Timestamp* object from given *ts*.
        :raises ValueError: The timestamp is invalid, e.g. missing or
            invalid timestamp component.
        """
        # If inadvertently passed a Timestamp object, use replace()
        # to create a clone.
        if isinstance(ts, cls):
            return ts.replace()

        if not isinstance(ts, str):
            raise TypeError(
                f'ts argument must be a string or a Timestamp object,'
                f' not {type(ts).__name__}')

        if len(ts) == 8 and not strict:
            # year, month and day are given only
            ts += '000000'
        return cls._from_mw(ts)

    def isoformat(self, sep: str = 'T') -> str:  # type: ignore[override]
        """Convert object to an ISO 8601 timestamp accepted by MediaWiki.

        datetime.datetime.isoformat does not postfix the ISO formatted
        date with a 'Z' unless a timezone is included, which causes
        MediaWiki ~1.19 and earlier to fail.
        """
        return self.strftime(self._ISO8601Format(sep))

    def totimestampformat(self) -> str:
        """Convert object to a MediaWiki internal timestamp."""
        return self.strftime(self.mediawikiTSFormat)

    def posix_timestamp(self) -> float:
        """Convert object to a POSIX timestamp.

        See Note in datetime.timestamp().

        .. versionadded:: 7.5
        """
        return self.replace(tzinfo=datetime.timezone.utc).timestamp()

    def posix_timestamp_format(self) -> str:
        """Convert object to a POSIX timestamp format.

        .. versionadded:: 7.5
        """
        return f'{self.posix_timestamp():.6f}'

    def __repr__(self) -> str:
        """Unify repr string between CPython and Pypy (T325905).

        .. versionadded:: 8.0
        """
        s = super().__repr__()
        return f'{type(self).__name__}{s[s.find("("):]}'

    def __str__(self) -> str:
        """Return a string format recognized by the API."""
        return self.isoformat()

    @classmethod
    def nowutc(cls, *, with_tz: bool = True) -> Timestamp:
        """Return the current UTC date and time.

        If *with_tz* is True it returns an aware :class:`Timestamp`
        object with UTC timezone by calling ``now(datetime.UTC)``. As
        such this is just a short way to get it.

        Otherwise the UTC timestamp is returned as a naive Timestamp
        object with timezone of None. This is equal to the
        :meth:`Timestamp.utcnow`.

        .. warning::
           Because naive datetime objects are treated by many datetime
           methods as local times, it is preferred to use aware
           Timestamps or datetimes to represent times in UTC. As such,
           it is not recommended to set *with_tz* to False.

        .. caution::
           You cannot compare, add or subtract offset-naive and offset-
           aware Timestamps/datetimes (i.e. missing or having timezone).
           A TypeError will be raised in such cases.

        .. versionadded:: 9.0
        .. seealso::
           - :python:`datetime.now()
             <library/datetime.html#datetime.datetime.now>`
           - :python:`datetime.utcnow()
             <library/datetime.html#datetime.datetime.utcnow>`
           - :python:`datetime.UTC<library/datetime.html#datetime.UTC>`

        :param with_tz: Whether to include UTC timezone or not
        """
        if with_tz:
            # datetime.UTC can only be used with Python 3.11+
            return cls.now(datetime.timezone.utc)

        ts = _time.time()
        gmt = _time.gmtime(ts)
        us = round(ts % 1 * 1e6)
        return cls(*gmt[:6], us)

    @classmethod
    def utcnow(cls) -> Timestamp:
        """Return the current UTC date and time, with `tzinfo` ``None``.

        This is like :meth:`Timestamp.now`, but returns the current UTC
        date and time, as a naive :class:`Timestamp` object. An aware
        current UTC datetime can be obtained by calling
        :meth:`Timestamp.nowutc`.

        .. note:: This method is deprecated since Python 3.12 but held
           here for backward compatibility because ``utcnow`` is widely
           used inside the framework to compare MediaWiki timestamps
           which are UTC-based. Neither :python:`datetime.fromisoformat()
           <library/datetime.html#datetime.datetime.fromisoformat>`
           implementations of Python < 3.11 nor :class:`Timestamp`
           specific :meth:`fromISOformat` supports timezone.

        .. warning::
           Because naive datetime objects are treated by many datetime
           methods as local times, it is preferred to use aware
           Timestamps or datetimes to represent times in UTC. As such,
           the recommended way to create an object representing the
           current time in UTC is by calling :meth:`Timestamp.nowutc`.

        .. hint::
           This method might be deprecated later.

        .. versionadded:: 9.0
        .. seealso::
           :python:`datetime.utcnow()
           <library/datetime.html#datetime.datetime.utcnow>`
        """
        if PYTHON_VERSION < (3, 12):
            return super().utcnow()
        return cls.nowutc(with_tz=False)


class TZoneFixedOffset(datetime.tzinfo):

    """Class building tzinfo objects for fixed-offset time zones.

    :param offset: a number indicating fixed offset in minutes east from
        UTC
    :param name: a string with name of the timezone
    """

    def __init__(self, offset: int, name: str) -> None:
        """Initializer."""
        self._offset = datetime.timedelta(minutes=offset)
        self._name = name

    def utcoffset(self, dt: datetime.datetime | None) -> datetime.timedelta:
        """Return the offset to UTC."""
        return self._offset

    def tzname(self, dt: datetime.datetime | None) -> str:
        """Return the name of the timezone."""
        return self._name

    def dst(self, dt: datetime.datetime | None) -> datetime.timedelta:
        """Return no daylight savings time."""
        return datetime.timedelta(0)

    def __repr__(self) -> str:
        """Return the internal representation of the timezone."""
        return (f'{type(self).__name__}'
                f'({self._offset.days * 86400 + self._offset.seconds}, '
                f'{self._name})')


def str2timedelta(
    string: str,
    timestamp: datetime.datetime | None = None,
) -> datetime.timedelta:
    """Return a timedelta for a shorthand duration.

    :param string: a string defining a time period:

    Examples::

        300s - 300 seconds
        36h - 36 hours
        7d - 7 days
        2w - 2 weeks (14 days)
        1y - 1 year

    :param timestamp: a timestamp to calculate a more accurate duration offset
        used by years
    :type timestamp: datetime.datetime
    :return: the corresponding timedelta object
    """
    key, duration = parse_duration(string)

    if key == 'y':  # years
        days = math.ceil(duration * 365.25)
        duration *= 12

        if timestamp:
            return pywikibot.date.apply_month_delta(
                timestamp.date(), month_delta=duration) - timestamp.date()
        return datetime.timedelta(days=days)

    # days, seconds, hours, weeks
    return datetime.timedelta(**{MW_KEYS[key]: duration})


def parse_duration(string: str) -> tuple[str, int]:
    """Return the key and duration extracted from the string.

    :param string: a string defining a time period

    Examples::

        300s - 300 seconds
        36h - 36 hours
        7d - 7 days
        2w - 2 weeks (14 days)
        1y - 1 year

    :return: key and duration extracted form the string
    """
    if len(string) < 2:
        raise ValueError('Time period should be a numeric value followed by '
                         'its qualifier')

    key, duration = string[-1], string[:-1]

    if key not in MW_KEYS:
        raise ValueError(f'Time period qualifier is unrecognized: {string}')
    if not duration.isdigit():
        raise ValueError(f"Time period's duration should be numeric: {string}")

    return key, int(duration)



"""XML reading module.

Each XmlEntry object represents a page, as read from an XML source

The XmlDump class reads a pages_current XML dump (like the ones offered on
https://dumps.wikimedia.org/backup-index.html) and offers a generator over
XmlEntry objects which can be used by other bots.

.. versionchanged:: 7.7
   *defusedxml* is used in favour of *xml.etree* if present to prevent
   vulnerable XML attacks. *defusedxml* 0.7.1 or higher is recommended.
"""
#
# (C) Pywikibot team, 2005-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import re
from dataclasses import dataclass
from typing import NamedTuple
from xml.etree.ElementTree import Element


try:
    from defusedxml.ElementTree import ParseError, iterparse
except ImportError:
    from xml.etree.ElementTree import iterparse, ParseError

from pywikibot.backports import Callable, Iterator
from pywikibot.tools import (
    ModuleDeprecationWrapper,
    issue_deprecation_warning,
    open_archive,
)


@dataclass
class XmlEntry:

    """Represent a page."""

    # TODO: there are more tags we can read.
    title: str
    ns: str
    id: str
    text: str
    username: str
    ipedit: bool
    timestamp: str
    editRestriction: str  # noqa: N815
    moveRestriction: str  # noqa: N815
    revisionid: str
    comment: str
    isredirect: bool


class Headers(NamedTuple):

    """Represent the common info of a page.

    .. versionadded:: 9.0
    """

    title: str
    ns: str
    pageid: str
    isredirect: bool
    edit_restriction: str
    move_restriction: str


class RawRev(NamedTuple):

    """Represent a raw revision.

    .. versionadded:: 9.0
    """

    headers: Headers
    revision: Element
    revid: int


class XmlDump:

    """Represents an XML dump file.

    Reads the local file at initialization,
    parses it, and offers access to the resulting XmlEntries via a generator.

    .. versionadded:: 7.2
       the `on_error` parameter
    .. versionchanged:: 7.2
       `allrevisions` parameter must be given as keyword parameter
    .. versionchanged:: 9.0
       `allrevisions` parameter is deprecated due to :phab:`T340804`,
       `revisions` parameter was introduced as replacement.
       `root` attribute was removed.

    Usage example:

    >>> from pywikibot import xmlreader
    >>> name = 'tests/data/xml/article-pear.xml'
    >>> dump = xmlreader.XmlDump(name, revisions='all')
    >>> for elem in dump.parse():
    ...     print(elem.title, elem.revisionid)
    ...
    ...
    Pear 185185
    Pear 185241
    Pear 185408
    Pear 188924
    >>>

    :param allrevisions: boolean
        If True, parse all revisions instead of only the latest one.
        Default: False.
    :param on_error: a callable which is invoked within :meth:`parse`
        method when a ParseError occurs. The exception is passed to this
        callable. Otherwise the exception is raised.
    :param revisions: which of four methods to use to parse the dump:
        * `first_found` (whichever revision is the first element)
        * `latest` (most recent revision, by largest `revisionid`)
        * `earliest` (first revision, by smallest `revisionid`)
        * `all` (all revisions for each page)
        Default: `first_found`
    """

    def __init__(
        self,
        filename,
        *,
        allrevisions: bool | str | None = None,
        # when allrevisions removed, revisions can default to 'latest'
        revisions: str = 'first_found',
        on_error: Callable[[ParseError], None] | None = None,
    ) -> None:
        """Initializer."""
        self.filename = filename
        self.on_error = on_error

        self.rev_actions = {
            'first_found': self._parse_only_first_found,
            'latest': self._parse_only_latest,
            'earliest': self._parse_only_earliest,
            'all': self._parse_all,
        }

        if allrevisions:
            issue_deprecation_warning(
                'allrevisions=True',
                "revisions='all'",
                since='9.0.0')
            revisions = 'all'
        elif revisions == 'first_found':
            issue_deprecation_warning(
                'allrevisions=False returns first revision found,'
                " usually earliest. For most recent, use revisions='latest'. "
                "For oldest, use revisions='earliest',"
                "'allrevisions'",
                since='9.0.0')

        if revisions not in self.rev_actions:
            actions = str(list(self.rev_actions.keys())).strip('[]')
            raise ValueError(f"'revisions' must be one of {actions}.")

        self._parse = self.rev_actions[revisions]
        self.uri = None

    def parse(self) -> Iterator[XmlEntry]:
        """Generator using ElementTree iterparse function.

        .. versionchanged:: 7.2
           if a ParseError occurs it can be handled by the callable
           given with `on_error` parameter of this instance.
        """
        with open_archive(self.filename) as source:
            context = iterparse(source, events=('start', 'end', 'start-ns'))
            root = None

            while True:
                try:
                    event, elem = next(context)
                except StopIteration:
                    return
                except ParseError as e:
                    if self.on_error:
                        self.on_error(e)
                        continue
                    raise

                if event == 'start-ns' and elem[0] == '':
                    self.uri = f'{{{elem[1]}}}'
                    continue

                # get the root element
                if event == 'start' and root is None:
                    root = elem

                if not (event == 'end' and elem.tag == f'{self.uri}page'):
                    continue

                yield from self._parse(elem)

                # clear references in the root, to allow garbage collection.
                elem.clear()
                root.clear()

    def _parse_only_first_found(self, elem: Element) -> Iterator[XmlEntry]:
        """Parser that yields the first revision found.

        .. versionadded:: 9.0
        """
        raw_revs = self._fetch_revs(elem)
        try:
            raw_rev = next(raw_revs)
            yield self._create_revision(raw_rev.headers, raw_rev.revision)
        except StopIteration:
            return

    def _parse_only_latest(self, elem: Element) -> Iterator[XmlEntry]:
        """Parser that yields only the latest revision."""
        raw_revs = self._fetch_revs(elem, with_id=True)
        raw_rev = max(raw_revs, default=None, key=lambda rev: rev.revid)
        if raw_rev is not None:
            yield self._create_revision(raw_rev.headers, raw_rev.revision)

    def _parse_only_earliest(self, elem: Element) -> Iterator[XmlEntry]:
        """Parser that yields only the earliest revision.

        .. versionadded:: 9.0
        """
        raw_revs = self._fetch_revs(elem, with_id=True)
        raw_rev = min(raw_revs, default=None, key=lambda rev: rev.revid)
        if raw_rev is not None:
            yield self._create_revision(raw_rev.headers, raw_rev.revision)

    def _parse_all(self, elem: Element) -> Iterator[XmlEntry]:
        """Parser that yields all revisions."""
        raw_revs = self._fetch_revs(elem)
        for raw_rev in raw_revs:
            yield self._create_revision(raw_rev.headers, raw_rev.revision)

    def _fetch_revs(self, elem: Element, with_id=False) -> Iterator[RawRev]:
        """Yield all revisions in a page.

        .. versionadded:: 9.0
        """
        uri = self.uri
        headers = self._headers(elem)
        for revision in elem.findall(f'{uri}revision'):
            revid = int(revision.findtext(f'{uri}id')) if with_id else 0
            yield RawRev(headers, revision, revid)

    @staticmethod
    def parse_restrictions(restrictions: str) -> tuple[str | None, str | None]:
        """Parse the characters within a restrictions tag.

        Returns strings representing user groups allowed to edit and
        to move a page, where None means there are no restrictions.

        .. versionadded:: 9.0
           replaces deprecated ``parseRestrictions`` function.
        """
        if not restrictions:
            return None, None

        edit_restriction, move_restriction = None, None

        edit_lock_match = re.search('edit=([^:]*)', restrictions)
        if edit_lock_match:
            edit_restriction = edit_lock_match[1]

        move_lock_match = re.search('move=([^:]*)', restrictions)
        if move_lock_match:
            move_restriction = move_lock_match[1]

        if restrictions == 'sysop':
            edit_restriction = 'sysop'
            move_restriction = 'sysop'

        return edit_restriction, move_restriction

    def _headers(self, elem: Element) -> Headers:
        """Extract headers from XML chunk."""
        uri = self.uri
        edit_restriction, move_restriction = self.parse_restrictions(
            elem.findtext(f'{uri}restrictions')
        )

        return Headers(
            title=elem.findtext(f'{uri}title'),
            ns=elem.findtext(f'{uri}ns'),
            pageid=elem.findtext(f'{uri}id'),
            isredirect=elem.findtext(f'{uri}redirect') is not None,
            edit_restriction=edit_restriction,
            move_restriction=move_restriction,
        )

    def _create_revision(
            self, headers: Headers, revision: Element
    ) -> XmlEntry:
        """Create a Single revision."""
        uri = self.uri
        contributor = revision.find(f'{uri}contributor')
        ip_editor = contributor.findtext(f'{uri}ip')
        username = ip_editor or contributor.findtext(f'{uri}username')
        username = username or ''  # username might be deleted

        return XmlEntry(
            title=headers.title,
            ns=headers.ns,
            id=headers.pageid,
            editRestriction=headers.edit_restriction,
            moveRestriction=headers.move_restriction,
            isredirect=headers.isredirect,
            text=revision.findtext(f'{uri}text'),
            username=username,
            ipedit=bool(ip_editor),
            timestamp=revision.findtext(f'{uri}timestamp'),
            revisionid=revision.findtext(f'{uri}id'),
            comment=revision.findtext(f'{uri}comment'),
            # could get comment, minor as well
        )


wrapper = ModuleDeprecationWrapper(__name__)
wrapper.add_deprecated_attr(
    'parseRestrictions',
    XmlDump.parse_restrictions,
    replacement_name='pywikibot.xmlreader.XmlDump.parseRestrictions',
    since='9.0.0')



"""Objects representing Namespaces of MediaWiki site."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from abc import ABCMeta
from collections.abc import Iterable, Mapping
from enum import IntEnum
from typing import Union

from pywikibot.backports import Iterable as IterableType
from pywikibot.tools import ComparableMixin, classproperty


SingleNamespaceType = Union[int, str, 'Namespace']
NamespaceArgType = Union[
    SingleNamespaceType,
    IterableType[SingleNamespaceType],
    None,
]


class BuiltinNamespace(IntEnum):

    """Builtin namespace enum."""

    MEDIA = -2
    SPECIAL = -1
    MAIN = 0
    TALK = 1
    USER = 2
    USER_TALK = 3
    PROJECT = 4
    PROJECT_TALK = 5
    FILE = 6
    FILE_TALK = 7
    MEDIAWIKI = 8
    MEDIAWIKI_TALK = 9
    TEMPLATE = 10
    TEMPLATE_TALK = 11
    HELP = 12
    HELP_TALK = 13
    CATEGORY = 14
    CATEGORY_TALK = 15

    @property
    def canonical(self) -> str:
        """Canonical form of MediaWiki built-in namespace.

        .. versionadded:: 7.1
        """
        name = '' if self == 0 else self.name.capitalize().replace('_', ' ')
        return name.replace('Mediawiki', 'MediaWiki')


class MetaNamespace(ABCMeta):

    """Metaclass for Namespace attribute settings.

    .. versionadded:: 9.0
    """

    def __new__(cls, name, bases, dic):
        """Set Namespace.FOO to BuiltinNamespace.FOO for each builtin ns."""
        dic.update(BuiltinNamespace.__members__)
        return super().__new__(cls, name, bases, dic)


class Namespace(Iterable, ComparableMixin, metaclass=MetaNamespace):

    """Namespace site data object.

    This is backwards compatible with the structure of entries
    in site._namespaces which were a list of::

        [customised namespace,
         canonical namespace name?,
         namespace alias*]

    If the canonical_name is not provided for a namespace between -2
    and 15, the MediaWiki built-in names are used.
    Image and File are aliases of each other by default.

    If only one of canonical_name and custom_name are available, both
    properties will have the same value.

    .. versionchanged:: 9.0
       metaclass from :class:`MetaNamespace`
    """

    # Hints of BuiltinNamespace types added with initializer
    MEDIA: int
    SPECIAL: int
    MAIN: int
    TALK: int
    USER: int
    USER_TALK: int
    PROJECT: int
    PROJECT_TALK: int
    FILE: int
    FILE_TALK: int
    MEDIAWIKI: int
    MEDIAWIKI_TALK: int
    TEMPLATE: int
    TEMPLATE_TALK: int
    HELP: int
    HELP_TALK: int
    CATEGORY: int
    CATEGORY_TALK: int

    def __init__(self, id,
                 canonical_name: str | None = None,
                 custom_name: str | None = None,
                 aliases: list[str] | None = None,
                 **kwargs) -> None:
        """Initializer.

        :param canonical_name: Canonical name
        :param custom_name: Name defined in server LocalSettings.php
        :param aliases: Aliases
        """
        self.id = id
        canonical_name = canonical_name or BuiltinNamespace(self.id).canonical

        assert custom_name is not None or canonical_name is not None, \
            'Namespace needs to have at least one name'

        self.custom_name = custom_name \
            if custom_name is not None else canonical_name
        self.canonical_name = canonical_name \
            if canonical_name is not None else custom_name

        if aliases:
            self.aliases = aliases
        elif id in (BuiltinNamespace.FILE, BuiltinNamespace.FILE_TALK):
            alias = 'Image'
            if id == BuiltinNamespace.FILE_TALK:
                alias += ' talk'
            self.aliases = [alias]
        else:
            self.aliases = []

        for key, value in kwargs.items():
            setattr(self, key, value)

    @classproperty
    def canonical_namespaces(cls) -> dict[int, str]:
        """Return the canonical forms of MediaWiki built-in namespaces.

        .. versionchanged:: 7.1
           implemented as classproperty using BuiltinNamespace IntEnum.
        """
        return {item.value: item.canonical for item in BuiltinNamespace}

    def _distinct(self):
        if self.custom_name == self.canonical_name:
            return [self.canonical_name, *self.aliases]
        return [self.custom_name, self.canonical_name, *self.aliases]

    def _contains_lowercase_name(self, name):
        """Determine a lowercase normalised name is a name of this namespace.

        :rtype: bool
        """
        return name in (x.lower() for x in self._distinct())

    def __contains__(self, item: str) -> bool:
        """Determine if item is a name of this namespace.

        The comparison is case insensitive, and item may have a single
        colon on one or both sides of the name.

        :param item: name to check
        """
        if item == '' and self.id == BuiltinNamespace.MAIN:
            return True

        name = Namespace.normalize_name(item)
        if not name:
            return False

        return self._contains_lowercase_name(name.lower())

    def __bool__(self) -> bool:
        """Obtain boolean method for Namespace class.

        This method is implemented to be independent from __len__ method.

        .. versionadded:: 7.0

        :return: Always return True like generic object class.
        """
        return True

    def __len__(self) -> int:
        """Obtain length of the iterable."""
        return len(self._distinct())

    def __iter__(self):
        """Return an iterator."""
        return iter(self._distinct())

    def __getitem__(self, index):
        """Obtain an item from the iterable."""
        if self.custom_name != self.canonical_name:
            if index == 0:
                return self.custom_name
            index -= 1

        return self.canonical_name if index == 0 else self.aliases[index - 1]

    @staticmethod
    def _colons(id, name):
        """Return the name with required colons, depending on the ID."""
        if id == BuiltinNamespace.MAIN:
            return ':'

        if id in (BuiltinNamespace.FILE, BuiltinNamespace.CATEGORY):
            return ':' + name + ':'

        return name + ':'

    def __str__(self) -> str:
        """Return the canonical string representation."""
        return self.canonical_prefix()

    def canonical_prefix(self):
        """Return the canonical name with required colons."""
        return Namespace._colons(self.id, self.canonical_name)

    def custom_prefix(self):
        """Return the custom name with required colons."""
        return Namespace._colons(self.id, self.custom_name)

    def __int__(self) -> int:
        """Return the namespace id."""
        return self.id

    def __index__(self) -> int:
        """Return the namespace id."""
        return self.id

    def __hash__(self):
        """Return the namespace id."""
        return self.id

    def __eq__(self, other):
        """Compare whether two namespace objects are equal."""
        if isinstance(other, int):
            return self.id == other

        if isinstance(other, Namespace):
            return self.id == other.id

        if isinstance(other, str):
            return other in self

        return False

    def __ne__(self, other):
        """Compare whether two namespace objects are not equal."""
        return not self.__eq__(other)

    def __mod__(self, other):
        """Apply modulo on the namespace id."""
        return self.id.__mod__(other)

    def __sub__(self, other):
        """Apply subtraction on the namespace id."""
        return self.id - other

    def __add__(self, other):
        """Apply addition on the namespace id."""
        return self.id + other

    def _cmpkey(self):
        """Return the ID as a comparison key."""
        return self.id

    def __repr__(self) -> str:
        """Return a reconstructable representation."""
        standard_attr = ['id', 'custom_name', 'canonical_name', 'aliases']
        extra = [(key, self.__dict__[key])
                 for key in sorted(self.__dict__)
                 if key not in standard_attr]

        if extra:
            kwargs = ', ' + ', '.join(
                key + f'={value!r}' for key, value in extra)
        else:
            kwargs = ''

        return (f'{self.__class__.__name__}(id={self.id}, '
                f'custom_name={self.custom_name!r}, '
                f'canonical_name={self.canonical_name!r}, '
                f'aliases={self.aliases!r}{kwargs})')

    @staticmethod
    def default_case(id, default_case=None):
        """Return the default fixed case value for the namespace ID."""
        # https://www.mediawiki.org/wiki/Manual:$wgCapitalLinkOverrides#Warning
        if id in (BuiltinNamespace.SPECIAL,
                  BuiltinNamespace.USER, BuiltinNamespace.USER_TALK,
                  BuiltinNamespace.MEDIAWIKI, BuiltinNamespace.MEDIAWIKI_TALK,
                  ):
            return 'first-letter'

        return default_case

    @classmethod
    def builtin_namespaces(cls, case: str = 'first-letter'):
        """Return a dict of the builtin namespaces."""
        return {e.value: cls(e.value, case=cls.default_case(e.value, case))
                for e in BuiltinNamespace}

    @staticmethod
    def normalize_name(name):
        """Remove an optional colon before and after name.

        TODO: reject illegal characters.
        """
        if name == '':
            return ''

        name = name.replace('_', ' ')
        parts = name.split(':', 4)
        count = len(parts)
        if count > 3 or (count == 3 and parts[2]):
            return False

        # Discard leading colon
        if count >= 2 and not parts[0] and parts[1]:
            return parts[1].strip()

        if parts[0]:
            return parts[0].strip()

        return False


class NamespacesDict(Mapping):

    """An immutable dictionary containing the Namespace instances."""

    def __init__(self, namespaces) -> None:
        """Create new dict using the given namespaces."""
        super().__init__()
        self._namespaces = namespaces
        self._namespace_names = {}
        for namespace in self._namespaces.values():
            for name in namespace:
                self._namespace_names[name.lower()] = namespace

    def __iter__(self):
        """Iterate over all namespaces."""
        return iter(self._namespaces)

    def __getitem__(self, key: Namespace | int | str) -> Namespace:
        """Get the namespace with the given key.

        :param key: namespace key
        """
        if isinstance(key, (Namespace, int)):
            try:
                return self._namespaces[key]
            except KeyError:
                raise KeyError(f'{key} is not a known namespace. Maybe you'
                               ' should clear the api cache.')

        namespace = self.lookup_name(key)
        if namespace:
            return namespace

        return super().__getitem__(key)

    def __getattr__(self, attr: Namespace | int | str) -> Namespace:
        """Get the namespace with the given key.

        :param attr: namespace key
        """
        # lookup_name access _namespaces
        if attr.isupper():
            if attr == 'MAIN':
                return self[0]

            namespace = self.lookup_name(attr)
            if namespace:
                return namespace

        return self.__getattribute__(attr)

    def __len__(self) -> int:
        """Get the number of namespaces."""
        return len(self._namespaces)

    def lookup_name(self, name: str) -> Namespace | None:
        """Find the Namespace for a name also checking aliases.

        :param name: Name of the namespace.
        """
        name = Namespace.normalize_name(name)
        if name is False:
            return None
        return self.lookup_normalized_name(name.lower())

    def lookup_normalized_name(self, name: str) -> Namespace | None:
        """Find the Namespace for a name also checking aliases.

        The name has to be normalized and must be lower case.

        :param name: Name of the namespace.
        """
        return self._namespace_names.get(name)

    def resolve(self, identifiers) -> list[Namespace]:
        """Resolve namespace identifiers to obtain Namespace objects.

        Identifiers may be any value for which int() produces a valid
        namespace id, except bool, or any string which
        Namespace.lookup_name successfully finds. A numerical string is
        resolved as an integer.

        :param identifiers: namespace identifiers
        :type identifiers: iterable of str or Namespace key, or a single
            instance of those types
        :return: list of Namespace objects in the same order as the
            identifiers
        :raises KeyError: a namespace identifier was not resolved
        :raises TypeError: a namespace identifier has an inappropriate
            type such as NoneType or bool
        """
        if isinstance(identifiers, (str, Namespace)):
            identifiers = [identifiers]
        else:
            # convert non-iterators to single item list
            try:
                iter(identifiers)
            except TypeError:
                identifiers = [identifiers]

        # lookup namespace names, and assume anything else is a key.
        # int(None) raises TypeError; however, bool needs special handling.
        namespaces = self._namespaces
        result = [NotImplemented if isinstance(ns, bool)
                  else self._lookup_name(ns)
                  if isinstance(ns, str) and not ns.lstrip('-').isdigit()
                  else namespaces.get(int(ns))
                  for ns in identifiers]

        if NotImplemented in result:
            raise TypeError(
                f'identifiers contains inappropriate types: {identifiers!r}'
            )

        # Namespace.lookup_name returns None if the name is not recognised
        if None in result:
            raise KeyError(
                'Namespace identifier(s) not recognised: {}'
                .format(','.join(str(identifier)
                                 for identifier, ns in zip(identifiers, result)
                                 if ns is None)))

        return result

    def _lookup_name(self, name):
        name = Namespace.normalize_name(name)
        if name is False:
            return None
        name = name.lower()

        for namespace in self._namespaces.values():
            if namespace._contains_lowercase_name(name):
                return namespace

        return None



#!/usr/bin/env python3
"""This script generates a family file from a given URL.

This script must be invoked with the pwb wrapper script/code entry point.

Usage::

    pwb generate_family_file.py [<url>] [<name>] [<dointerwiki>] [<verify>]

Parameters are optional. They must be given consecutively but may be
omitted if there is no successor parameter. The parameters are::

    <url>:         an url from where the family settings are loaded
    <name>:        the family name without "_family.py" tail.
    <dointerwiki>: predefined answer (y|s|n) to add multiple site codes
    <verify>:      disable certificate validaton `(y|n)

Example::

    pwb generate_family_file.py https://www.mywiki.bogus/wiki/Main_Page mywiki

This will create the file mywiki_family.py in families folder of your
base directory.

.. versionchanged:: 7.0
   moved to pywikibot.scripts folder; create family files in families
   folder of your base directory instead of pywikibot/families.
.. versionchanged:: 8.1
   [s]trict can be given for <dointerwiki> parameter to ensure that
   sites are from the given domain.
.. versionchanged:: 8.4
   If the url scheme is missing, ``https`` will be used.
"""
#
# (C) Pywikibot team, 2010-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import re
import string
import sys
from contextlib import suppress
from pathlib import Path
from textwrap import fill
from urllib.parse import urlparse, urlunparse


# see pywikibot.family.py
# Legal characters for Family name and Family langs keys
NAME_CHARACTERS = string.ascii_letters + string.digits
# nds_nl code alias requires "_"n
# dash must be the last char to be reused as regex
CODE_CHARACTERS = string.ascii_lowercase + string.digits + '_-'


class FamilyFileGenerator:

    """Family file creator object."""

    def __init__(self,
                 url: str | None = None,
                 name: str | None = None,
                 dointerwiki: str | None = None,
                 verify: str | None = None) -> None:
        """Parameters are optional. If missing the script asks for the values.

        :param url: an url from where the family settings are loaded
        :param name: the family name without "_family.py" tail.
        :param dointerwiki: Predefined answer to add multiple site
            codes. Pass `Y` or `y` for yes, `S` or `s` for strict which
            only includes site of the same domain (usually for Wikimedia
            sites), `N` or `n` for no and `E` or `e` if you want to edit
            the collection of sites.
        :param verify: If a certificate verification fails, you may pass
            `Y` or `y` to disable certificate validaton `N` or `n` to
            keep it enabled.
        """
        from pywikibot.scripts import _import_with_no_user_config

        # from pywikibot.site_detect import MWSite and
        # from pywikibot.config import base_dir
        # when required but disable user-config checks
        # so the family can be created first,
        # and then used when generating the user-config
        self.Wiki = _import_with_no_user_config(
            'pywikibot.site_detect').site_detect.MWSite
        self.base_dir = _import_with_no_user_config(
            'pywikibot.config').config.base_dir

        self.base_url = url
        self.name = name
        self.dointerwiki = dointerwiki
        self.verify = verify

        self.wikis = {}  # {'https://wiki/$1': Wiki('https://wiki/$1'), ...}
        self.langs = []  # [Wiki('https://wiki/$1'), ...]

    @staticmethod
    def show(*args, **kwargs):
        """Wrapper around print to be mocked in tests."""
        print(*args, **kwargs)

    def get_params(self) -> bool:
        """Ask for parameters if necessary."""
        if self.base_url is None:
            with suppress(KeyboardInterrupt):
                url = input('Please insert URL to wiki: ')
            if not url:
                return False
            url = urlparse(url, 'https')
            if not url.netloc and url.path:
                self.base_url = urlunparse((url.scheme, url.path, url.netloc,
                                            *url[3:]))
            else:
                self.base_url = urlunparse(url)

        if self.name is None:
            with suppress(KeyboardInterrupt):
                self.name = input('Please insert a short name (eg: freeciv): ')
            if not self.name:
                return False

        if any(x not in NAME_CHARACTERS for x in self.name):
            self.show(f'ERROR: Name of family "{self.name}" must be ASCII'
                      ' letters  and digits [a-zA-Z0-9]')
            return False

        return True

    def get_wiki(self):
        """Get wiki from base_url."""
        import pywikibot
        from pywikibot.exceptions import FatalServerError
        pywikibot.info('Generating family file from ' + self.base_url)
        for verify in (True, False):
            try:
                w = self.Wiki(self.base_url, verify=verify)
            except FatalServerError:
                pywikibot.exception()
                pywikibot.info()
                if not pywikibot.bot.input_yn(
                    'Retry with disabled ssl certificate validation',
                    default=self.verify, automatic_quit=False,
                        force=self.verify is not None):
                    break
            else:
                return w, verify
        return None, None

    def run(self) -> None:
        """Main method, generate family file."""
        if not self.get_params():
            return

        w, verify = self.get_wiki()
        if w is None:
            return

        self.wikis[w.lang] = w
        self.show('\n=================================='
                  f'\nAPI url: {w.api}'
                  f'\nMediaWiki version: {w.version}'
                  '\n==================================\n')

        self.getlangs(w)
        self.getapis()
        self.writefile(verify)

    def getlangs(self, w) -> None:
        """Determine site code of a family.

        .. versionchanged:: 8.1
           with [e]dit the interwiki list can be given delimited by
           space or comma or both. With [s]trict only sites with the
           same domain are collected. A [h]elp answer was added to show
           more information about possible answers.
        """
        self.show('Determining other sites...', end='')
        try:
            self.langs = w.langs
            self.show(fill(' '.join(sorted(wiki['prefix']
                                           for wiki in self.langs))))
        except Exception as e:
            self.langs = []
            self.show(e, '; continuing...')

        if len([lang for lang in self.langs if lang['url'] == w.iwpath]) == 0:
            if w.private_wiki:
                w.lang = self.name
            self.langs.append({'language': w.lang,
                               'local': '',
                               'prefix': w.lang,
                               'url': w.iwpath})

        code_len = len(self.langs)
        if code_len > 1:
            if self.dointerwiki is None:
                while True:
                    makeiw = input(
                        '\n'
                        f'There are {code_len} sites available.'
                        ' Do you want to generate interwiki links?\n'
                        'This might take a long time. '
                        '([y]es, [s]trict, [N]o, [e]dit), [h]elp) ').lower()
                    if makeiw in ('y', 's', 'n', 'e', ''):
                        break
                    self.show(
                        '\n'
                        '[y]es:    create interwiki links for all sites\n'
                        '[s]trict: yes, but for sites with same domain only\n'
                        '[N]o:     no, use the current site only (default)\n'
                        '[e]dit:   get a list delimited with space or comma\n'
                        '[h]elp:   this help message'
                    )
            else:
                makeiw = self.dointerwiki

            if makeiw in ('n', ''):
                self.langs = [wiki for wiki in self.langs
                              if wiki['url'] == w.iwpath]
            elif makeiw == 's':
                domain = '.'.join(urlparse(w.server).hostname.split('.')[1:])
                self.langs = [wiki for wiki in self.langs
                              if domain in wiki['url']]

            elif makeiw == 'e':
                for wiki in self.langs:
                    self.show(wiki['prefix'], wiki['url'])
                do_langs = re.split(' *,| +',
                                    input('Which sites do you want: '))
                self.langs = [wiki for wiki in self.langs
                              if wiki['prefix'] in do_langs
                              or wiki['url'] == w.iwpath]

        for wiki in self.langs:
            assert all(x in CODE_CHARACTERS for x in wiki['prefix']), \
                'Family {} code {} must be ASCII lowercase ' \
                'letters and digits [a-z0-9] or underscore/dash [_-]' \
                .format(self.name, wiki['prefix'])

    def getapis(self) -> None:
        """Load other site pages."""
        self.show(f'Loading {len(self.langs)} wikis... ')
        remove = []
        for lang in self.langs:
            key = lang['prefix']
            self.show(f'  * {key}... ', end='')
            if key not in self.wikis:
                try:
                    self.wikis[key] = self.Wiki(lang['url'])
                    self.show('downloaded')
                except Exception as e:
                    self.show(e)
                    remove.append(lang)
            else:
                self.show('in cache')

        for lang in remove:
            self.langs.remove(lang)

    def writefile(self, verify) -> None:
        """Write the family file."""
        fp = Path(self.base_dir, 'families', f'{self.name}_family.py')
        self.show(f'Writing {fp}... ')

        if fp.exists() and input(
                f'{fp} already exists. Overwrite? (y/n) ').lower() == 'n':
            self.show('Terminating.')
            sys.exit(1)

        code_hostname_pairs = '\n        '.join(
            f"'{k}': '{urlparse(w.server).netloc}',"
            for k, w in self.wikis.items())

        code_path_pairs = '\n            '.join(
            f"'{k}': '{w.scriptpath}',"
            for k, w in self.wikis.items())

        code_protocol_pairs = '\n            '.join(
            f"'{k}': '{urlparse(w.server).scheme}',"
            for k, w in self.wikis.items())

        content = family_template % {
            'url': self.base_url, 'name': self.name,
            'code_hostname_pairs': code_hostname_pairs,
            'code_path_pairs': code_path_pairs,
            'code_protocol_pairs': code_protocol_pairs}
        if not verify:
            # assuming this is the same for all codes
            content += """

    def verify_SSL_certificate(self, code: str) -> bool:
        return False
"""
        fp.parent.mkdir(parents=True, exist_ok=True)
        fp.write_text(content, encoding='utf-8')


family_template = """\
\"\"\"
This family file was auto-generated by generate_family_file.py script.

Configuration parameters:
  url = %(url)s
  name = %(name)s

Please do not commit this to the Git repository!
\"\"\"
from pywikibot import family


class Family(family.Family):  # noqa: D101

    name = '%(name)s'
    langs = {
        %(code_hostname_pairs)s
    }

    def scriptpath(self, code):
        return {
            %(code_path_pairs)s
        }[code]

    def protocol(self, code):
        return {
            %(code_protocol_pairs)s
        }[code]
"""


def main() -> None:
    """Process command line arguments and generate a family file."""
    if len(sys.argv) > 1 and sys.argv[1] == '-help':
        print(__doc__)
    else:
        FamilyFileGenerator(*sys.argv[1:]).run()


if __name__ == '__main__':
    main()



"""Object representing a MediaWiki category page."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from collections import defaultdict
from typing import Any

import pywikibot
from pywikibot.backports import Generator
from pywikibot.page._page import Page


__all__ = ('Category', )


class Category(Page):

    """A page in the Category: namespace."""

    def __init__(self, source, title: str = '', sort_key=None) -> None:
        """Initializer.

        All parameters are the same as for Page() Initializer.
        """
        self.sortKey = sort_key
        super().__init__(source, title, ns=14)
        if self.namespace() != 14:
            raise ValueError(
                f"'{self.title()}' is not in the category namespace!")

    def aslink(self, sort_key: str | None = None) -> str:
        """Return a link to place a page in this Category.

        .. warning:: Use this only to generate a "true" category link,
           not for interwikis or text links to category pages.

        **Usage:**

        >>> site = pywikibot.Site('wikipedia:test')
        >>> cat = pywikibot.Category(site, 'Foo')
        >>> cat.aslink()
        '[[Category:Foo]]'
        >>> cat = pywikibot.Category(site, 'Foo', sort_key='bar')
        >>> cat.aslink()
        '[[Category:Foo|bar]]'
        >>> cat.aslink('baz')
        '[[Category:Foo|baz]]'

        :param sort_key: The sort key for the article to be placed in this
            Category; if omitted, default sort key is used.
        """
        key = sort_key or self.sortKey
        title_with_sort_key = self.title(with_section=False)
        if key is not None:
            title_with_sort_key += '|' + key
        return f'[[{title_with_sort_key}]]'

    def subcategories(self, *,
                      recurse: int | bool = False,
                      **kwargs: Any) -> Generator[Category, None, None]:
        """Yield all subcategories of the current category.

        **Usage:**

        >>> site = pywikibot.Site('wikipedia:en')
        >>> cat = pywikibot.Category(site, 'Contents')
        >>> next(cat.subcategories())
        Category('Category:Wikipedia administration')
        >>> len(list(cat.subcategories(recurse=2, total=50)))
        50

        Subcategories of the same level of each subtree are yielded
        first before the next subcategories level are yielded. For example
        having this category tree:

        .. code-block:: text

           A
           +-- B
           |   +-- E
           |   |   +-- H
           |   +-- F
           |   +-- G
           +-- C
           |   +-- I
           |   |   +-- E
           |   |       +-- H
           |   +-- J
           |       +-- K
           |       +-- L
           |           +-- G
           +-- D

        Subcategories are yields in the following order:
        *B, C, D, E, F, G, H, I, J, E, H, K, L, G*

        .. seealso:: :attr:`categoryinfo`
        .. warning:: Categories may have infinite recursions of
           subcategories. If ``recurse`` option is given as ``True`` or
           an ``int`` value and this value is less than
           `sys.getrecursionlimit()`, an ``RecursionError`` may be
           raised. Be careful if passing this generator to a collection
           in such case.
        .. versionchanged:: 8.0
           all parameters are keyword arguments only. Additional
           parameters are supported. The order of subcategories are
           yielded was changed. The old order was
           *B, E, H, F, G, C, I, E, H, J, K, L, G, D*

        :param recurse: if not False or 0, also iterate articles in
            subcategories. If an int, limit recursion to this number of
            levels. (Example: ``recurse=1`` will iterate articles in
            first-level subcats, but no deeper.)
        :param kwargs: Additional parameters. Refer to
            :meth:`APISite.categorymembers()
            <pywikibot.site._generators.GeneratorsMixin.categorymembers>`
            for complete list (*member_type* excluded).
        """
        if kwargs.pop('member_type', False):
            raise TypeError('subcategories() got an unexpected keyword '
                            "argument 'member_type'")

        if not self.categoryinfo['subcats']:
            return

        if not isinstance(recurse, bool) and recurse:
            recurse -= 1

        yield from self.members(member_type='subcat', recurse=recurse,
                                **kwargs)

    def articles(self, *,
                 recurse: int | bool = False,
                 total: int | None = None,
                 **kwargs: Any) -> Generator[Page, None, None]:
        """Yield all articles in the current category.

        Yields all pages in the category that are not subcategories.
        Duplicates are filtered. To enable duplicates use :meth:`members`
        with ``member_type=['page', 'file']`` instead.

        **Usage:**

        >>> site = pywikibot.Site('wikipedia:test')
        >>> cat = pywikibot.Category(site, 'Pywikibot')
        >>> list(cat.articles())
        [Page('Pywikibot nobots test')]
        >>> for p in cat.articles(recurse=1, namespaces=2, total=3):
        ...     print(p.depth)
        ...
        2
        3
        4

        .. warning:: Categories may have infinite recursions of
           subcategories. If ``recurse`` option is given as ``True`` or
           an ``int`` value and this value is less than
           `sys.getrecursionlimit()`, an ``RecursionError`` may be
           raised. Be careful if passing this generator to a collection
           in such case.
        .. versionchanged:: 8.0
           all parameters are keyword arguments only.

        :param recurse: if not False or 0, also iterate articles in
            subcategories. If an int, limit recursion to this number of
            levels. (Example: ``recurse=1`` will iterate articles in
            first-level subcats, but no deeper.)
        :param total: iterate no more than this number of pages in
            total (at all levels)
        :param kwargs: Additional parameters. Refer to
            :meth:`APISite.categorymembers()
            <pywikibot.site._generators.GeneratorsMixin.categorymembers>`
            for complete list (*member_type* excluded).
        """
        if kwargs.pop('member_type', False):
            raise TypeError(
                "articles() got an unexpected keyword argument 'member_type'")

        member_type = ['page', 'file']
        if not recurse:
            yield from self.members(
                member_type=member_type, total=total, **kwargs)
            return

        seen = set()
        for member in self.members(
                member_type=member_type, recurse=recurse, **kwargs):
            if member.pageid in seen:
                continue

            seen.add(member.pageid)
            yield member
            if total is not None:
                total -= 1
                if total == 0:
                    return

    def members(self, *,
                recurse: int | bool = False,
                total: int | None = None,
                **kwargs: Any) -> Generator[Page, None, None]:
        """Yield all category contents (subcats, pages, and files).

        **Usage:**

        >>> site = pywikibot.Site('wikipedia:test')
        >>> cat = pywikibot.Category(site, 'Pywikibot')
        >>> list(cat.members(member_type='subcat'))
        [Category('Category:Subpage testing')]
        >>> list(cat.members(member_type=['page', 'file']))
        [Page('Pywikibot nobots test')]

        Calling this method with ``member_type='subcat'`` is equal to
        calling :meth:`subcategories`. Calling this method with
        ``member_type=['page', 'file']`` is equal to calling
        :meth:`articles` except that the later will filter duplicates.

        .. seealso:: :meth:`APISite.categorymembers()
           <pywikibot.site._generators.GeneratorsMixin.categorymembers>`
        .. warning:: Categories may have infinite recursions of
           subcategories. If ``recurse`` option is given as ``True`` or
           an ``int`` value and this value is less than
           `sys.getrecursionlimit()`, an ``RecursionError`` may be
           raised. Be careful if passing this generator to a collection
           in such case.
        .. versionchanged:: 8.0
           all parameters are keyword arguments only. Additional
           parameters are supported.

        :param recurse: if not False or 0, also iterate articles in
            subcategories. If an int, limit recursion to this number of
            levels. (Example: ``recurse=1`` will iterate articles in
            first-level subcats, but no deeper.)
        :param total: iterate no more than this number of pages in
            total (at all levels)
        :param kwargs: Additional parameters. Refer to
            :meth:`APISite.categorymembers()
            <pywikibot.site._generators.GeneratorsMixin.categorymembers>`
            for complete list.
        """
        for member in self.site.categorymembers(self, total=total, **kwargs):
            yield member
            if total is not None:
                total -= 1
                if total == 0:
                    return

        if not recurse:
            return

        if not isinstance(recurse, bool):
            recurse -= 1

        for subcat in self.subcategories():
            for member in subcat.members(
                    recurse=recurse, total=total, **kwargs):
                yield member
                if total is not None:
                    total -= 1
                    if total == 0:
                        return

    def isEmptyCategory(self) -> bool:  # noqa: N802
        """Return True if category has no members (including subcategories)."""
        ci = self.categoryinfo
        return sum(ci[k] for k in ['files', 'pages', 'subcats']) == 0

    def isHiddenCategory(self) -> bool:  # noqa: N802
        """Return True if the category is hidden."""
        return 'hiddencat' in self.properties()

    @property
    def categoryinfo(self) -> dict[str, Any]:
        """Return a dict containing information about the category.

        The dict contains values for numbers of pages, subcategories,
        files, and total contents.

        .. seealso:: :meth:`APISite.categoryinfo()
           <pywikibot.site._apisite.APISite.categoryinfo>`
        """
        return self.site.categoryinfo(self)

    def newest_pages(
        self,
        total: int | None = None
    ) -> Generator[Page, None, None]:
        """Return pages in a category ordered by the creation date.

        If two or more pages are created at the same time, the pages are
        returned in the order they were added to the category. The most
        recently added page is returned first.

        It only allows to return the pages ordered from newest to oldest, as it
        is impossible to determine the oldest page in a category without
        checking all pages. But it is possible to check the category in order
        with the newly added first and it yields all pages which were created
        after the currently checked page was added (and thus there is no page
        created after any of the cached but added before the currently
        checked).

        :param total: The total number of pages queried.
        :return: A page generator of all pages in a category ordered by the
            creation date. From newest to oldest.

            .. note:: It currently only returns Page instances and not a
               subclass of it if possible. This might change so don't
               expect to only get Page instances.
        """
        def check_cache(latest):
            """Return the cached pages in order and not more than total."""
            cached = []
            for timestamp in sorted((ts for ts in cache if ts > latest),
                                    reverse=True):
                # The complete list can be removed, it'll either yield all of
                # them, or only a portion but will skip the rest anyway
                cached += cache.pop(timestamp)[:None if total is None else
                                               total - len(cached)]
                if total and len(cached) >= total:
                    break  # already got enough
            assert total is None or len(cached) <= total, \
                'Number of caches is more than total number requested'
            return cached

        # all pages which have been checked but where created before the
        # current page was added, at some point they will be created after
        # the current page was added. It saves all pages via the creation
        # timestamp. Be prepared for multiple pages.
        cache = defaultdict(list)
        # TODO: Make site.categorymembers is usable as it returns pages
        # There is no total defined, as it's not known how many pages need to
        # be checked before the total amount of new pages was found. In worst
        # case all pages of a category need to be checked.
        for member in pywikibot.data.api.QueryGenerator(
            site=self.site, parameters={
                'list': 'categorymembers', 'cmsort': 'timestamp',
                'cmdir': 'older', 'cmprop': 'timestamp|title',
                'cmtitle': self.title()}):
            # TODO: Upcast to suitable class
            page = pywikibot.Page(self.site, member['title'])
            assert page.namespace() == member['ns'], \
                'Namespace of the page is not consistent'
            cached = check_cache(pywikibot.Timestamp.fromISOformat(
                member['timestamp']))
            yield from cached
            if total is not None:
                total -= len(cached)
                if total <= 0:
                    break
            cache[page.oldest_revision.timestamp].append(page)
        else:
            # clear cache
            assert total is None or total > 0, \
                'As many items as given in total already returned'
            yield from check_cache(pywikibot.Timestamp.min)



"""Miscellaneous helper functions (not wiki-dependent)."""
#
# (C) Pywikibot team, 2008-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import abc
import hashlib
import importlib.metadata
import ipaddress
import os
import re
import stat
import subprocess
from contextlib import suppress
from functools import total_ordering, wraps
from types import TracebackType
from typing import Any
from warnings import catch_warnings, showwarning, warn

import packaging.version

import pywikibot  # T306760
from pywikibot.backports import PYTHON_VERSION, SPHINX_RUNNING, Callable
from pywikibot.tools._deprecate import (
    ModuleDeprecationWrapper,
    add_decorated_full_name,
    add_full_name,
    deprecate_arg,
    deprecated,
    deprecated_args,
    deprecated_signature,
    get_wrapper_depth,
    issue_deprecation_warning,
    manage_wrapping,
    redirect_func,
    remove_last_args,
)
from pywikibot.tools._unidata import _first_upper_exception


try:  # Python 3.14+
    from compression import bz2, gzip, lzma
except ModuleNotFoundError:
    import bz2
    import gzip
    import lzma


__all__ = (
    # deprecating functions
    'ModuleDeprecationWrapper',
    'add_decorated_full_name',
    'add_full_name',
    'deprecate_arg',
    'deprecated',
    'deprecated_args',
    'deprecated_signature',
    'get_wrapper_depth',
    'issue_deprecation_warning',
    'manage_wrapping',
    'redirect_func',
    'remove_last_args',

    # other tools
    'PYTHON_VERSION',
    'SPHINX_RUNNING',
    'as_filename',
    'is_ip_address',
    'is_ip_network',
    'has_module',
    'classproperty',
    'suppress_warnings',
    'ComparableMixin',
    'first_lower',
    'first_upper',
    'strtobool',
    'normalize_username',
    'MediaWikiVersion',
    'open_archive',
    'merge_unique_dicts',
    'file_mode_checker',
    'compute_file_hash',
    'cached',
)


def is_ip_address(value: str) -> bool:
    """Check if a value is a valid IPv4 or IPv6 address.

    .. versionadded:: 6.1
       Was renamed from ``is_IP()``.
    .. seealso:: :func:`is_ip_network`

    :param value: value to check
    """
    with suppress(ValueError):
        ipaddress.ip_address(value)
        return True

    return False


def is_ip_network(value: str) -> bool:
    """Check if a value is a valid range of IPv4 or IPv6 addresses.

    .. versionadded:: 9.0
    .. seealso:: :func:`is_ip_address`

    :param value: value to check
    """
    with suppress(ValueError):
        ipaddress.ip_network(value)
        return True

    return False


def has_module(module: str, version: str | None = None) -> bool:
    """Check if a module can be imported.

    .. versionadded:: 3.0

    .. versionchanged:: 6.1
       Dependency of distutils was dropped because the package will be
       removed with Python 3.12.
    """
    try:
        metadata_version = importlib.metadata.version(module)
    except importlib.metadata.PackageNotFoundError:
        return False
    if version:

        required_version = packaging.version.Version(version)
        module_version = packaging.version.Version(metadata_version)

        if module_version < required_version:
            warn(f'Module version {module_version} is lower than requested '
                 f'version {required_version}', ImportWarning, stacklevel=2)
            return False

    return True


class classproperty:  # noqa: N801

    """Descriptor class to access a class method as a property.

    This class may be used as a decorator::

        class Foo:

            _bar = 'baz'  # a class property

            @classproperty
            def bar(cls):  # a class property method
                return cls._bar

    Foo.bar gives 'baz'.

    .. versionadded:: 3.0
    """

    def __init__(self, cls_method) -> None:
        """Initializer: hold the class method and documentation."""
        self.method = cls_method
        self.__annotations__ = self.method.__annotations__
        doc = self.method.__doc__
        self.__doc__ = f':class:`classproperty<tools.classproperty>` {doc}'

        rtype = self.__annotations__.get('return')
        if rtype:
            lines = doc.splitlines()

            if len(lines) > 2 and PYTHON_VERSION < (3, 13):
                spaces = ' ' * re.search('[^ ]', lines[2]).start()
            else:
                spaces = ''

            self.__doc__ += f'\n\n{spaces}:rtype: {rtype}'

    def __get__(self, instance, owner):
        """Get the attribute of the owner class by its method."""
        if SPHINX_RUNNING:
            return self

        return self.method(owner)


class suppress_warnings(catch_warnings):  # noqa: N801

    """A decorator/context manager that temporarily suppresses warnings.

    Those suppressed warnings that do not match the parameters will be
    raised shown upon exit.

    .. versionadded:: 3.0
    """

    def __init__(
        self,
        message: str = '',
        category=Warning,
        filename: str = ''
    ) -> None:
        """Initialize the object.

        The parameter semantics are similar to those of
        `warnings.filterwarnings`.

        :param message: A string containing a regular expression that
            the start of the warning message must match.
            (case-insensitive)
        :param category: A class (a subclass of Warning) of which the
            warning category must be a subclass in order to match.
        :type category: type
        :param filename: A string containing a regular expression that
            the start of the path to the warning module must match.
            (case-sensitive)
        """
        self.message_match = re.compile(message, re.IGNORECASE).match
        self.category = category
        self.filename_match = re.compile(filename).match
        super().__init__(record=True)

    def __enter__(self) -> None:
        """Catch all warnings and store them in `self.log`."""
        self.log = super().__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None
    ) -> None:
        """Stop logging warnings and show those that do not match to params."""
        super().__exit__(exc_type, exc_val, exc_tb)
        for warning in self.log:
            if (
                not issubclass(warning.category, self.category)
                or not self.message_match(str(warning.message))
                or not self.filename_match(warning.filename)
            ):
                showwarning(
                    warning.message, warning.category, warning.filename,
                    warning.lineno, warning.file, warning.line)

    def __call__(self, func):
        """Decorate func to suppress warnings."""
        @wraps(func)
        def suppressed_func(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return suppressed_func


# From http://python3porting.com/preparing.html
class ComparableMixin(abc.ABC):

    """Mixin class to allow comparing to other objects which are comparable.

    .. versionadded:: 3.0
    """

    @abc.abstractmethod
    def _cmpkey(self) -> Any:
        """Abstract method to return key for comparison of objects.

        This ensures that ``_cmpkey`` method is defined in subclass.

        .. versionadded:: 8.1.2
        """

    def __lt__(self, other):
        """Compare if self is less than other."""
        return other > self._cmpkey()

    def __le__(self, other):
        """Compare if self is less equals other."""
        return other >= self._cmpkey()

    def __eq__(self, other):
        """Compare if self is equal to other."""
        return other == self._cmpkey()

    def __ge__(self, other):
        """Compare if self is greater equals other."""
        return other <= self._cmpkey()

    def __gt__(self, other):
        """Compare if self is greater than other."""
        return other < self._cmpkey()

    def __ne__(self, other):
        """Compare if self is not equal to other."""
        return other != self._cmpkey()


def first_lower(string: str) -> str:
    """Return a string with the first character uncapitalized.

    Empty strings are supported. The original string is not changed.

    **Example**:

    >>> first_lower('Hello World')
    'hello World'

    .. versionadded:: 3.0
    """
    return string[:1].lower() + string[1:]


def first_upper(string: str) -> str:
    """Return a string with the first character capitalized.

    Empty strings are supported. The original string is not changed.

    **Example**:

    >>> first_upper('hello World')
    'Hello World'

    .. versionadded:: 3.0
    .. note:: MediaWiki doesn't capitalize some characters the same way
       as Python. This function tries to be close to MediaWiki's
       capitalize function in title.php. See :phab:`T179115` and
       :phab:`T200357`.
    """
    first = string[:1]
    return (_first_upper_exception(first) or first.upper()) + string[1:]


def as_filename(string: str, repl: str = '_') -> str:
    r"""Return a string with characters are valid for filenames.

    Replace characters that are not possible in file names on some
    systems, but still are valid in MediaWiki titles:

        - Unix: ``/``
        - MediaWiki: ``/:\\``
        - Windows: ``/:\\"?*``

    Spaces are possible on most systems, but are bad for URLs.

    **Example**:

    >>> as_filename('How are you?')
    'How_are_you_'
    >>> as_filename('Say: "Hello"')
    'Say___Hello_'
    >>> as_filename('foo*bar', '')
    'foobar'
    >>> as_filename('foo', 'bar')
    Traceback (most recent call last):
    ...
    ValueError: Invalid repl parameter 'bar'
    >>> as_filename('foo', '?')
    Traceback (most recent call last):
    ...
    ValueError: Invalid repl parameter '?'

    .. versionadded:: 8.0

    :param string: the string to be modified
    :param repl: the replacement character
    :raises ValueError: Invalid repl parameter
    """
    pattern = r':*?/\\" '
    if len(repl) > 1 or len(repl) == 1 and repl in pattern:
        raise ValueError(f'Invalid repl parameter {repl!r}')
    return re.sub(f'[{pattern}]', repl, string)


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to True or False.

    This is a reimplementation of distutils.util.strtobool due to
    :pep:`632#Migration Advice`

    **Example**:

    >>> strtobool('yes')
    True
    >>> strtobool('Off')
    False
    >>> strtobool('aye')
    Traceback (most recent call last):
    ...
    ValueError: invalid truth value 'aye'

    .. versionadded:: 7.1

    :param val: True values are 'y', 'yes', 't', 'true', 'on', and '1';
        false values are 'n', 'no', 'f', 'false', 'off', and '0'.
    :raises ValueError: `val` is not a valid truth value
    """
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    if val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    raise ValueError(f'invalid truth value {val!r}')


def normalize_username(username) -> str | None:
    """Normalize the username.

    .. versionadded:: 3.0
    """
    if not username:
        return None
    username = re.sub('[_ ]+', ' ', username).strip()
    return first_upper(username)


@total_ordering
class MediaWikiVersion:

    """Version object to allow comparing 'wmf' versions with normal ones.

    The version mainly consist of digits separated by periods. After
    that is a suffix which may only be 'wmf<number>', 'alpha',
    'beta<number>' or '-rc.<number>' (the - and . are optional). They
    are considered from old to new in that order with a version number
    without suffix is considered the newest. This secondary difference
    is stored in an internal _dev_version attribute.

    Two versions are equal if their normal version and dev version are
    equal. A version is greater if the normal version or dev version is
    greater. For example::

        1.34 < 1.34.1 < 1.35wmf1 < 1.35alpha < 1.35beta1 < 1.35beta2
        < 1.35-rc-1 < 1.35-rc.2 < 1.35

    Any other suffixes are considered invalid.

    .. versionadded:: 3.0

    .. versionchanged:: 6.1
       Dependency of distutils was dropped because the package will be
       removed with Python 3.12.
    """

    MEDIAWIKI_VERSION = re.compile(
        r'(\d+(?:\.\d+)+)(-?wmf\.?(\d+)|alpha|beta(\d+)|-?rc\.?(\d+)|.*)?')

    def __init__(self, version_str: str) -> None:
        """Initializer.

        :param version_str: version to parse
        """
        self._parse(version_str)

    def _parse(self, version_str: str) -> None:
        version_match = MediaWikiVersion.MEDIAWIKI_VERSION.fullmatch(
            version_str)

        if not version_match:
            raise ValueError(f'Invalid version number "{version_str}"')

        components = [int(n) for n in version_match[1].split('.')]

        # The _dev_version numbering scheme might change. E.g. if a stage
        # between 'alpha' and 'beta' is added, 'beta', 'rc' and stable releases
        # are reassigned (beta=3, rc=4, stable=5).

        if version_match[3]:  # wmf version
            self._dev_version = (0, int(version_match[3]))
        elif version_match[4]:
            self._dev_version = (2, int(version_match[4]))
        elif version_match[5]:
            self._dev_version = (3, int(version_match[5]))
        elif version_match[2] in ('alpha', '-alpha'):
            self._dev_version = (1, )
        else:
            for handled in ('wmf', 'alpha', 'beta', 'rc'):
                # if any of those pops up here our parser has failed
                assert handled not in version_match[2], \
                    f'Found "{handled}" in "{version_match[2]}"'
            if version_match[2]:
                pywikibot.logging.debug(
                    'Additional unused version part {version_match[2]!r}')
            self._dev_version = (4, )

        self.suffix = version_match[2] or ''
        self.version = tuple(components)

    @staticmethod
    def from_generator(generator: str) -> MediaWikiVersion:
        """Create instance from a site's generator attribute."""
        prefix = 'MediaWiki '

        if not generator.startswith(prefix):
            raise ValueError(f'Generator string ({generator!r}) must start '
                             f'with "{prefix}"')

        return MediaWikiVersion(generator[len(prefix):])

    def __str__(self) -> str:
        """Return version number with optional suffix."""
        return '.'.join(str(v) for v in self.version) + self.suffix

    def __repr__(self) -> str:
        """Return version number representation, mainly used by tests.

        .. versionadded:: 10.0
        """
        return f"'{self}'"

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, str):
            other = MediaWikiVersion(other)
        elif not isinstance(other, MediaWikiVersion):
            return False

        return self.version == other.version and \
            self._dev_version == other._dev_version

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, str):
            other = MediaWikiVersion(other)
        elif not isinstance(other, MediaWikiVersion):
            raise TypeError(f"Comparison between 'MediaWikiVersion' and "
                            f"'{type(other).__name__}' unsupported")

        if self.version != other.version:
            return self.version < other.version
        return self._dev_version < other._dev_version


def open_archive(filename: str, mode: str = 'rb', use_extension: bool = True):
    """Open a file and uncompress it if needed.

    This function supports bzip2, gzip, 7zip, lzma, and xz as
    compression containers. It uses the packages available in the
    standard library for bzip2, gzip, lzma, and xz so they are always
    available. 7zip is only available when a 7za program is available
    and only supports reading from it.

    The compression is either selected via the magic number or file
    ending.

    .. versionadded:: 3.0

    :param filename: The filename.
    :param use_extension: Use the file extension instead of the magic
        number to determine the type of compression (default True). Must
        be True when writing or appending.
    :param mode: The mode in which the file should be opened. It may
        either be 'r', 'rb', 'a', 'ab', 'w' or 'wb'. All modes open the
        file in binary mode. It defaults to 'rb'.
    :raises ValueError: When 7za is not available or the opening mode is
        unknown or it tries to write a 7z archive.
    :raises FileNotFoundError: When the filename doesn't exist and it
        tries to read from it or it tries to determine the compression
        algorithm.
    :raises OSError: When it's not a 7z archive but the file extension
        is 7z. It is also raised by bz2 when its content is invalid.
        gzip does not immediately raise that error but only on reading
        it.
    :raises lzma.LZMAError: When error occurs during compression or
        decompression or when initializing the state with lzma or xz.
    :return: A file-like object returning the uncompressed data in
        binary mode.
    :rtype: file-like object
    """
    # extension_map maps magic_number to extension.
    # Unfortunately, legacy LZMA container has no magic number
    extension_map = {
        b'BZh': 'bz2',
        b'\x1F\x8B\x08': 'gz',
        b"7z\xBC\xAF'\x1C": '7z',
        b'\xFD7zXZ\x00': 'xz',
    }

    if mode in ('r', 'a', 'w'):
        mode += 'b'
    elif mode not in ('rb', 'ab', 'wb'):
        raise ValueError(f'Invalid mode: "{mode}"')

    if use_extension:
        # if '.' not in filename, it'll be 1 character long but otherwise
        # contain the period
        extension = filename[filename.rfind('.'):][1:]
    else:
        if mode != 'rb':
            raise ValueError('Magic number detection only when reading')
        with open(filename, 'rb') as f:
            magic_number = f.read(8)

        for pattern, ext in extension_map.items():
            if magic_number.startswith(pattern):
                extension = ext
                break
        else:
            extension = ''

    if extension == 'bz2':
        binary = bz2.BZ2File(filename, mode)

    elif extension == 'gz':
        binary = gzip.open(filename, mode)

    elif extension == '7z':
        if mode != 'rb':
            raise NotImplementedError('It is not possible to write a 7z file.')

        try:
            process = subprocess.Popen(['7za', 'e', '-bd', '-so', filename],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE,
                                       bufsize=65535)
        except OSError:
            raise ValueError(
                f'7za is not installed or cannot uncompress "{filename}"')

        stderr = process.stderr.read()
        process.stderr.close()
        if stderr != b'':
            process.stdout.close()
            raise OSError(
                f'Unexpected STDERR output from 7za {stderr}')
        binary = process.stdout

    elif extension in ('lzma', 'xz'):
        lzma_fmts = {'lzma': lzma.FORMAT_ALONE, 'xz': lzma.FORMAT_XZ}
        binary = lzma.open(filename, mode, format=lzma_fmts[extension])

    else:  # assume it's an uncompressed file
        binary = open(filename, 'rb')

    return binary


def merge_unique_dicts(*args, **kwargs):
    """Return a merged dict and make sure that the original keys are unique.

    The positional arguments are the dictionaries to be merged. It is
    also possible to define an additional dict using the keyword
    arguments.

    .. versionadded:: 3.0
    """
    args = [*list(args), dict(kwargs)]
    conflicts = set()
    result = {}
    for arg in args:
        conflicts |= set(arg.keys()) & set(result.keys())
        result.update(arg)
    if conflicts:
        raise ValueError('Multiple dicts contain the same keys: {}'
                         .format(', '.join(sorted(str(key)
                                                  for key in conflicts))))
    return result


def file_mode_checker(
    filename: str | bytes | os.PathLike,
    mode: int = 0o600,
    quiet: bool = False,
    create: bool = False
) -> None:
    """Check file mode and update it, if needed.

    .. versionadded:: 3.0

    :param filename: filename path
    :param mode: requested file mode
    :param quiet: warn about file mode change if False.
    :param create: create the file if it does not exist already
    :raise IOError: The file does not exist and `create` is False.
    """
    try:
        st_mode = os.stat(filename).st_mode
    except OSError:  # file does not exist
        if not create:
            raise
        os.close(os.open(filename, os.O_CREAT | os.O_EXCL, mode))
        return

    warn_str = 'File {0} had {1:o} mode; converted to {2:o} mode.'
    if stat.S_ISREG(st_mode) and (st_mode - stat.S_IFREG != mode):
        os.chmod(filename, mode)
        # re-read and check changes
        if os.stat(filename).st_mode != st_mode and not quiet:
            warn(warn_str.format(filename, st_mode - stat.S_IFREG, mode),
                 stacklevel=2)


def compute_file_hash(filename: str | os.PathLike,
                      sha: str | Callable[[], Any] = 'sha1',
                      bytes_to_read: int | None = None) -> str:
    """Compute file hash.

    Result is expressed as hexdigest().

    .. versionadded:: 3.0
    .. versionchanged:: 8.2
       *sha* may be  also a hash constructor, or a callable that returns
       a hash object.


    :param filename: filename path
    :param sha: hash algorithm available with hashlib: ``sha1()``,
        ``sha224()``, ``sha256()``, ``sha384()``, ``sha512()``,
        ``blake2b()``, and ``blake2s()``. Additional algorithms like
        ``md5()``, ``sha3_224()``, ``sha3_256()``, ``sha3_384()``,
        ``sha3_512()``, ``shake_128()`` and ``shake_256()`` may also be
        available. *sha* must either be a hash algorithm name as a str
        like ``'sha1'`` (default), a hash constructor like
        ``hashlib.sha1``, or a callable that returns a hash object like
        ``lambda: hashlib.sha1()``.
    :param bytes_to_read: only the first bytes_to_read will be
        considered; if file size is smaller, the whole file will be
        considered.
    """
    with open(filename, 'rb') as f:
        if PYTHON_VERSION < (3, 11) or bytes_to_read is not None:
            digest = sha() if callable(sha) else hashlib.new(sha)
            size = os.path.getsize(filename)
            bytes_to_read = min(bytes_to_read or size, size)
            step = 1 << 20
            while bytes_to_read > 0:
                read_bytes = f.read(min(bytes_to_read, step))
                assert read_bytes  # make sure we actually read bytes
                bytes_to_read -= len(read_bytes)
                digest.update(read_bytes)
        else:
            digest = hashlib.file_digest(f, sha)

    return digest.hexdigest()


def cached(*arg: Callable) -> Any:
    """Decorator to cache information of an object.

    The wrapper adds an attribute to the instance which holds the result
    of the decorated method. The attribute's name is the method name
    with preleading underscore.

    Usage::

        @cached
        def this_method(self)

        @cached
        def that_method(self, force=False)

    No parameter may be used with this decorator. Only a force parameter
    may be used with the decorated method. All other parameters are
    discarded and lead to a TypeError.

    .. note:: A property must be decorated on top of the property method
       below other decorators. This decorator must not be used with
       functions.
    .. versionadded:: 7.3

    :raises TypeError: decorator must be used without arguments
    """
    fn = arg and arg[0]
    if not callable(fn):
        raise TypeError(
            '"cached" decorator must be used without arguments.') from None

    @wraps(fn)
    def wrapper(obj: object, *, force=False) -> Any:
        cache_name = '_' + fn.__name__
        if force:
            with suppress(AttributeError):
                delattr(obj, cache_name)
        try:
            return getattr(obj, cache_name)
        except AttributeError:
            val = fn(obj)
            setattr(obj, cache_name, val)
            return val

    return wrapper



"""Library to log the bot in to a wiki account."""
#
# (C) Pywikibot team, 2003-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import datetime
import os
import re
import webbrowser
from ast import literal_eval
from enum import IntEnum
from pathlib import Path
from textwrap import fill
from typing import Any
from warnings import warn

import pywikibot
from pywikibot import __url__, config
from pywikibot.comms import http
from pywikibot.exceptions import APIError, NoUsernameError
from pywikibot.tools import (
    PYTHON_VERSION,
    deprecated,
    file_mode_checker,
    normalize_username,
)


try:
    import mwoauth
except ImportError as e:
    mwoauth = e

TEST_RUNNING = os.environ.get('PYWIKIBOT_TEST_RUNNING', '0') == '1'


class _PasswordFileWarning(UserWarning):

    """The format of password file is incorrect."""


# On some wikis you are only allowed to run a bot if there is a link to
# the bot's user page in a specific list.
# If bots are listed in a template, the templates name must be given as
# second parameter, otherwise it must be None
botList = {
    'wikipedia': {
        'simple': ['Wikipedia:Bots', '/links']
    },
}


class LoginStatus(IntEnum):

    """Enum for Login statuses.

    >>> LoginStatus.NOT_ATTEMPTED
    LoginStatus(-3)
    >>> LoginStatus.IN_PROGRESS.value
    -2
    >>> LoginStatus.NOT_LOGGED_IN.name
    'NOT_LOGGED_IN'
    >>> int(LoginStatus.AS_USER)
    0
    >>> LoginStatus(-3).name
    'NOT_ATTEMPTED'
    >>> LoginStatus(0).name
    'AS_USER'
    """

    NOT_ATTEMPTED = -3
    IN_PROGRESS = -2
    NOT_LOGGED_IN = -1
    AS_USER = 0

    def __repr__(self) -> str:
        """Return internal representation."""
        return f'LoginStatus({self})'


class LoginManager:

    """Site login manager."""

    def __init__(self, password: str | None = None,
                 site: pywikibot.site.BaseSite | None = None,
                 user: str | None = None) -> None:
        """Initializer.

        All parameters default to defaults in user-config.

        :param site: Site object to log into
        :param user: username to use. If user is None, the username is
            loaded from config.usernames.
        :param password: password to use
        :raises pywikibot.exceptions.NoUsernameError: No username is
            configured for the requested site.
        """
        site = self.site = site or pywikibot.Site()
        if not user:
            config_names = config.usernames

            code_to_usr = config_names[site.family.name] or config_names['*']
            try:
                user = code_to_usr.get(site.code) or code_to_usr['*']
            except KeyError:
                raise NoUsernameError(
                    f'ERROR: username for {site.family.name}:{site.code} is'
                    ' undefined.\nIf you have a username for that site,'
                    ' please add a line to user config file (user-config.py)'
                    ' as follows:\n'
                    f"usernames['{site.family.name}']['{site.code}'] ="
                    " 'myUsername'"
                )
        self.password = password
        self.login_name = self.username = user
        if getattr(config, 'password_file', ''):
            self.readPassword()

    def __repr__(self) -> str:
        """Return representation string for LoginManager.

        .. versionadded:: 10.0
        """
        return f'{type(self).__name__}(user={self.username!r})'

    def check_user_exists(self) -> None:
        """Check that the username exists on the site.

        .. seealso:: :api:`Users`

        :raises pywikibot.exceptions.NoUsernameError: Username doesn't exist in
            user list.
        """
        # convert any Special:BotPassword usernames to main account equivalent
        main_username = self.username
        if '@' in self.username:
            warn('When using BotPasswords it is recommended that you store'
                 ' your login credentials in a password_file instead. See '
                 f'{__url__}/BotPasswords for instructions and more'
                 ' information.', stacklevel=2)
            main_username = self.username.partition('@')[0]

        try:
            data = self.site.allusers(start=main_username, total=1)
            user = next(data, {'name': None})
        except APIError as e:
            if e.code == 'readapidenied':
                pywikibot.warning(f"Could not check user '{main_username}' "
                                  f'exists on {self.site}')
                return
            raise

        if user['name'] != main_username:
            # Report the same error as server error code NotExists
            msg = f"Username '{main_username}' does not exist on {self.site}"
            raise NoUsernameError(msg)

    def botAllowed(self) -> bool:
        """Check whether the bot is listed on a specific page.

        This allows bots to comply with the policy on the respective
        wiki.
        """
        code, fam = self.site.code, self.site.family.name
        if code in botList.get(fam, []):
            botlist_pagetitle, bot_template_title = botList[fam][code]
            botlist_page = pywikibot.Page(self.site, botlist_pagetitle)
            if bot_template_title:
                for template, params in botlist_page.templatesWithParams():
                    if (template.title() == bot_template_title
                            and params[0] == self.username):
                        return True
            else:
                for linked_page in botlist_page.linkedPages():
                    if linked_page.title(with_ns=False) == self.username:
                        return True
            return False

        # No bot policies on other sites
        return True

    def login_to_site(self) -> None:
        """Login to the site."""
        # This is overridden in ClientLoginManager
        raise NotImplementedError

    def storecookiedata(self) -> None:
        """Store cookie data."""
        http.cookie_jar.save(ignore_discard=True)

    def readPassword(self) -> None:
        """Read passwords from a file.

        .. warning:: **Do not forget to remove read access for other
           users!** Use chmod 600 for password-file.

        All lines below should be valid Python tuples in the form
        ``(code, family, username, password)``,
        ``(family, username, password)`` or ``(username, password)`` to
        set a default password for an username. The last matching entry
        will be used, so default usernames should occur above specific
        usernames.

        .. note:: For BotPasswords the password should be given like a
           :class:`BotPassword` object.

        The file must be either encoded in ASCII or UTF-8.

        **Example**::

         ('my_username', 'my_default_password')
         ('wikipedia', 'my_wikipedia_user', 'my_wikipedia_pass')
         ('en', 'wikipedia', 'my_en_wikipedia_user', 'my_en_wikipedia_pass')
         ('my_username', BotPassword('my_suffix', 'my_password'))

        .. versionchanged:: 10.2
           Raises ValueError instead of AttributeError if password_file
               is not set.
        .. versionchanged:: 10.7.1
           Due to vulnerability issue the password lines are no longer
           evaluated as Python source but parsed as literals.
           Raises ValueError if an exception occurs while evaluating a
           password line.


        :raises ValueError: `password_file` is not set in the
            ``user-config.py`` or entries are critical malformed
        :raises FileNotFoundError: password file does not exist
        """
        if config.password_file is None:
            raise ValueError('password_file is not set in the user-config.py')

        # Set path to password file relative to the user_config
        # but fall back on absolute path for backwards compatibility
        password_path = Path(config.base_dir, config.password_file)

        params = {} if PYTHON_VERSION < (3, 13) else {'follow_symlinks': False}
        # test for symlink required for Python < 3.13
        if not password_path.is_file(**params) or password_path.is_symlink():
            password_path = Path(config.password_file)

        # ignore this check when running tests
        if os.environ.get('PYWIKIBOT_TEST_RUNNING', '0') == '0' \
           and (not password_path.is_file(**params)
                or password_path.is_symlink()):
            raise FileNotFoundError(
                f'Password file {password_path.name} does not exist in '
                f'{password_path.parent}'
            )

        # We fix password file permission first.
        file_mode_checker(password_path, mode=config.private_files_permission)

        lines = password_path.read_text('utf-8').splitlines()
        line_len = len(lines)
        for n, line in enumerate(reversed(lines)):
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            # sanity check
            if (line.count(', ') > 4 or len(line) > 250) and not TEST_RUNNING:
                raise ValueError('Password line too long or too complex')

            botpassword = 'BotPassword' in line
            if botpassword:
                line = line.replace('BotPassword', '')

            try:
                entry = literal_eval(line)
            except Exception:
                # catch the exceptions to not leak passwords to console or logs
                raise ValueError('Invalid password line format') from None

            if not isinstance(entry, tuple):
                warn(f'Invalid tuple in line {line_len - n}',
                     _PasswordFileWarning, stacklevel=2)
                continue

            if not 2 <= (entry_len := len(entry)) <= 4:
                warn(f'The length of tuple in line {line_len - n} should be 2 '
                     f'to 4, {entry_len} given ({entry})',
                     _PasswordFileWarning, stacklevel=2)
                continue

            if botpassword:
                entry = (*entry[:-1], BotPassword(*entry[-1]))

            code, family, username, password = (
                self.site.code, self.site.family.name)[:4 - entry_len] + entry

            if (normalize_username(username) == self.username
                    and family == self.site.family.name
                    and code == self.site.code):
                if isinstance(password, str):
                    self.password = password
                    break

                if isinstance(password, BotPassword):
                    self.password = password.password
                    self.login_name = password.login_name(self.username)
                    break

                warn('Invalid password format', _PasswordFileWarning,
                     stacklevel=2)

    def login(self, retry: bool = False, autocreate: bool = False) -> bool:
        """Attempt to log into the server.

        .. seealso:: :api:`Login`

        :param retry: infinitely retry if the API returns an unknown error
        :param autocreate: if true, allow auto-creation of the account
                           using unified login

        :raises pywikibot.exceptions.NoUsernameError: Username is not
            recognised by the site.
        """
        if not self.password:
            # First check that the username exists,
            # to avoid asking for a password that will not work.
            if not autocreate:
                self.check_user_exists()

            # As we don't want the password to appear on the screen, we set
            # password = True
            self.password = pywikibot.input(
                f'Password for user {self.login_name} on {self.site}'
                ' (no characters will be shown):',
                password=True
            )
        else:
            pywikibot.info(f'Logging in to {self.site} as {self.login_name}')

        try:
            self.login_to_site()
        except APIError as e:
            error_code = e.code

            if error_code in ('NotExists', 'Illegal', 'readapidenied',
                              'Failed', 'Aborted', 'FAIL'):
                error_msg = f'{e.code} on {self.site}: {e.info}'
                raise NoUsernameError(error_msg)

            pywikibot.error(f'Login failed ({error_code}).')
            if retry:
                self.password = None
                return self.login(retry=False)

        else:
            self.storecookiedata()
            pywikibot.log('Should be logged in now')
            return True

        return False


class ClientLoginManager(LoginManager):

    """Supply login_to_site method to use API interface.

    .. versionchanged:: 8.0
       2FA login was enabled. LoginManager was moved from :mod:`data.api`
       to :mod:`login` module and renamed to *ClientLoginManager*.
    .. versionchanged:: 10.2
       Secondary authentication via email was enabled.
    .. seealso::
       - https://www.mediawiki.org/wiki/Extension:OATHAuth
       - https://www.mediawiki.org/wiki/Extension:EmailAuth
    """

    # API login parameters mapping
    mapping = {
        'user': ('lgname', 'username'),
        'password': ('lgpassword', 'password'),
        'ldap': ('lgdomain', 'domain'),
        'token': ('lgtoken', 'logintoken'),
        'result': ('result', 'status'),
        'success': ('Success', 'PASS'),
        'fail': ('Failed', 'FAIL'),
        'reason': ('reason', 'message')
    }

    def keyword(self, key):
        """Get API keyword from mapping."""
        return self.mapping[key][self.action != 'login']

    def _login_parameters(self, *, botpassword: bool = False
                          ) -> dict[str, str]:
        """Return login parameters."""
        if botpassword:
            self.action = 'login'
        else:
            self.action = 'clientlogin'

        # prepare default login parameters
        parameters = {'action': self.action,
                      self.keyword('user'): self.login_name,
                      self.keyword('password'): self.password}

        if self.action == 'login':
            parameters['lgtoken'] = self.site.tokens['login']

        if self.action == 'clientlogin':
            # clientlogin requires non-empty loginreturnurl
            parameters['loginreturnurl'] = 'https://example.com'
            parameters['rememberMe'] = '1'
            parameters['logintoken'] = self.site.tokens['login']

        if self.site.family.ldapDomain:
            parameters[self.keyword('ldap')] = self.site.family.ldapDomain

        return parameters

    def login_to_site(self) -> None:
        """Login to the site.

        Note, this doesn't do anything with cookies. The http module
        takes care of all the cookie stuff. Throws exception on failure.

        .. versionchanged:: 8.0
           2FA login was implemented.
        .. versionchanged:: 10.2
           Secondary authentication via email was implemented.

        :raises RuntimeError: Unexpected API login response key or
            unexpected API login requests response
        :raises APIError: API login error
        """
        if hasattr(self, '_waituntil') \
           and datetime.datetime.now() < self._waituntil:
            diff = self._waituntil - datetime.datetime.now()
            pywikibot.warning(f'Too many tries, waiting {diff.seconds}'
                              ' seconds before retrying.')
            pywikibot.sleep(diff.seconds)

        self.site._loginstatus = LoginStatus.IN_PROGRESS

        # Bot passwords username contains @,
        # otherwise @ is not allowed in usernames.
        # @ in bot password is deprecated,
        # but we don't want to break bots using it.
        parameters = self._login_parameters(
            botpassword='@' in self.login_name or '@' in self.password)

        # base login request
        login_request = self.site._request(use_get=False,
                                           parameters=parameters)
        while True:
            # try to login
            try:
                login_result = login_request.submit()
            except pywikibot.exceptions.APIError as e:  # pragma: no cover
                login_result = {'error': e.__dict__}

            # clientlogin response can be clientlogin or error
            if self.action in login_result:
                response = login_result[self.action]
                result_key = self.keyword('result')
            elif 'error' in login_result:
                response = login_result['error']
                result_key = 'code'
            else:
                raise RuntimeError('Unexpected API login response key.')

            status = response[result_key]
            fail_reason = response.get(self.keyword('reason'), '')
            if status == self.keyword('success'):
                return

            if status in ('NeedToken', 'WrongToken', 'badtoken'):
                # if incorrect login token was used,
                # force relogin and generate fresh one
                pywikibot.error(f'{status}: Received incorrect login token.'
                                ' Forcing re-login.')
                # invalidate superior wiki cookies (T224712)
                pywikibot.data.api._invalidate_superior_cookies(
                    self.site.family)
                token = response.get('token')
                if not token:
                    self.site.tokens.clear()
                    token = self.site.tokens['login']
                login_request[self.keyword('token')] = token
                continue

            if status == 'UI':  # pragma: no cover
                # find token key
                token_key = None
                for request in response['requests']:
                    if 'fields' in request:
                        fields = request['fields']
                        keys = list(fields)
                        if len(keys) == 1:
                            token_key = keys[0]
                            field = fields[token_key]
                        break

                if not token_key:
                    raise RuntimeError('Unexpected API login requests response'
                                       f':\n{response["requests"]}')

                pywikibot.info(fill(response['message'], 77))
                if fail := response['messagecode'].endswith(('-failure',
                                                             '-failed')):
                    pywikibot.info(fill(field['help'], 77))
                token_val = pywikibot.input(field['label'], password=True)
                login_request[token_key] = token_val
                login_request['logincontinue'] = True

                if not fail:
                    del login_request['username']
                    del login_request['password']
                    del login_request['rememberMe']

                continue

            login_throttled = response.get('messagecode') == 'login-throttled'
            if (status == 'Throttled' or status == self.keyword('fail')
                    and (login_throttled or 'wait' in fail_reason)):
                wait = response.get('wait')
                if wait:
                    delta = datetime.timedelta(seconds=int(wait))
                else:
                    match = re.search(r'(\d+) (seconds|minutes)', fail_reason)
                    if match:
                        delta = datetime.timedelta(**{match[2]: int(match[1])})
                    else:
                        delta = datetime.timedelta()
                self._waituntil = datetime.datetime.now() + delta

            break

        if 'error' in login_result:
            raise pywikibot.exceptions.APIError(**response)

        raise pywikibot.exceptions.APIError(code=status, info=fail_reason)

    @deprecated("site.tokens['login']", since='8.0.0')
    def get_login_token(self) -> str | None:
        """Fetch login token.

        .. deprecated:: 8.0

        :return: login token
        """
        return self.site.tokens['login']


class BotPassword:

    """BotPassword object for storage in password file."""

    def __init__(self, suffix: str, password: str) -> None:
        """Initializer.

        BotPassword function by using a separate password paired with a
        suffixed username of the form <username>@<suffix>.

        :param suffix: Suffix of the login name
        :param password: bot password
        :raises _PasswordFileWarning: suffix improperly specified
        """
        if '@' in suffix:
            warn('The BotPassword entry should only include the suffix',
                 _PasswordFileWarning, stacklevel=2)
        self.suffix = suffix
        self.password = password

    def login_name(self, username: str) -> str:
        """Construct the login name from the username and suffix.

        :param username: username (without suffix)
        """
        return f'{username}@{self.suffix}'


class OauthLoginManager(LoginManager):

    """Site login manager using OAuth."""

    # NOTE: Currently OauthLoginManager use mwoauth directly to complete OAuth
    # authentication process

    def __init__(self, password: str | None = None,
                 site: pywikibot.site.BaseSite | None = None,
                 user: str | None = None) -> None:
        """Initializer.

        All parameters default to defaults in user-config.

        :param site: Site object to log into
        :param user: consumer key
        :param password: consumer secret
        :raises pywikibot.exceptions.NoUsernameError: No username is
            configured for the requested site.
        :raises ImportError: mwoauth isn't installed
        """
        if isinstance(mwoauth, ImportError):
            raise ImportError(f'mwoauth is not installed: {mwoauth}.')
        assert password is not None and user is not None
        super().__init__(password=None, site=site, user=None)
        if self.password:
            warn(
                f'Password exists in password file for {self.site}: '
                f'{self.username}. Password is unnecessary and should be'
                ' removed if OAuth enabled.',
                _PasswordFileWarning, stacklevel=2
            )
        self._consumer_token = (user, password)
        self._access_token: tuple[str, str] | None = None

    def login(self, retry: bool = False, force: bool = False) -> bool:
        """Attempt to log into the server.

        .. seealso:: :api:`Login`

        :param retry: infinitely retry if exception occurs during
            authentication.
        :param force: force to re-authenticate
        """
        if self.access_token is None or force:
            pywikibot.info(f'Logging in to {self.site} via OAuth consumer '
                           f'{self.consumer_token[0]}')
            consumer_token = mwoauth.ConsumerToken(*self.consumer_token)
            handshaker = mwoauth.Handshaker(
                self.site.base_url(self.site.path()), consumer_token)
            try:
                redirect, request_token = handshaker.initiate()
                pywikibot.stdout('Authenticate via web browser..')
                webbrowser.open(redirect)
                pywikibot.stdout(
                    'If your web browser does not open automatically, please '
                    f'point it to: {redirect}'
                )
                request_qs = pywikibot.input('Response query string: ')
                access_token = handshaker.complete(request_token, request_qs)
                self._access_token = (access_token.key, access_token.secret)
                return True
            except Exception as e:
                pywikibot.error(e)
                if retry:
                    return self.login(retry=True, force=force)
                return False
        else:
            pywikibot.info(f'Logged in to {self.site} via consumer '
                           f'{self.consumer_token[0]}')
            return True

    @property
    def consumer_token(self) -> tuple[str, str]:
        """Return OAuth consumer key token and secret token.

        .. seealso:: :api:`Tokens`
        """
        return self._consumer_token

    @property
    def access_token(self) -> tuple[str, str] | None:
        """OAuth access key token and secret token.

        .. seealso:: :api:`Tokens`

        :getter: Return OAuth access key token and secret token.
        :setter: Add OAuth access key token and secret token.
            Implemented to discard user interaction token fetching,
            usually for tests.

            .. versionadded:: 10.0
        """
        return self._access_token

    @access_token.setter
    def access_token(self, token: tuple[str, str]) -> None:
        """Add OAuth access key token and secret token."""
        self._access_token = token

    @property
    def identity(self) -> dict[str, Any] | None:
        """Get identifying information about a user via an authorized token.

        .. versionchanged:: 9.6
           *leeway* parameter for ``mwoauth.identify`` function was
           increased to 30.0 seconds.
        """
        if self.access_token is None:
            pywikibot.error('Access token not set')
            return None

        consumer_token = mwoauth.ConsumerToken(*self.consumer_token)
        access_token = mwoauth.AccessToken(*self.access_token)
        try:
            identity = mwoauth.identify(self.site.base_url(self.site.path()),
                                        consumer_token,
                                        access_token,
                                        leeway=30.0)
        except Exception as e:
            pywikibot.error(e)
        else:
            return identity

        return None



"""Object representing interface to toolforge tools.

.. versionadded:: 7.7
"""
#
# (C) Pywikibot team, 2022-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import collections
import re
from http import HTTPStatus
from warnings import warn

import pywikibot
from pywikibot.tools import deprecated, deprecated_args, remove_last_args


class WikiBlameMixin:

    """Page mixin for main authorship.

    .. versionadded:: 7.7
    """

    #: Supported wikipedia site codes
    WIKIBLAME_CODES = 'als', 'bar', 'de', 'en', 'it', 'nds', 'sco'

    def _check_wh_supported(self) -> None:
        """Check if WikiHistory is supported."""
        if self.site.family.name != 'wikipedia':
            raise NotImplementedError(
                'main_authors method is implemented for wikipedia family only')

        if (code := self.site.code) not in self.WIKIBLAME_CODES:
            raise NotImplementedError(
                f'main_authors method is not implemented for wikipedia:{code}')

        if (ns := self.namespace()) not in (0, 4, 10, 12, 14, 100):
            raise NotImplementedError(
                f'main_authors method is not implemented for {ns} namespace')

        if not self.exists():
            raise pywikibot.exceptions.NoPageError(self)

    @deprecated('authorsship', since='9.3.0')
    @deprecated_args(onlynew=None)  # since 9.2.0
    def main_authors(self) -> collections.Counter[str, int]:
        """Retrieve the 5 topmost main authors of an article.

        Sample:

        >>> import pywikibot
        >>> site = pywikibot.Site('wikipedia:de')
        >>> page = pywikibot.Page(site, 'Project:Pywikibot')
        >>> auth = page.main_authors()
        >>> auth.most_common(1)
        [('DrTrigon', 37)]

        .. deprecated:: 9.3
           use :meth:`authorship` instead.
        .. seealso:: :meth:`authorship` for further informations

        :return: Percentage of edits for each username

        :raise NotImplementedError: unsupported site or unsupported
            namespace.
        :raise NoPageError: The page does not exist.
        :raise TimeoutError: WikiHistory timeout
        """
        return collections.Counter(
            {user: int(cnt) for user, (_, cnt) in self.authorship(5).items()})

    @remove_last_args(['revid', 'date'])  # since 10.1.0
    def authorship(
        self,
        n: int | None = None,
        *,
        min_chars: int = 0,
        min_pct: float = 0.0,
        max_pct_sum: float | None = None,
    ) -> dict[str, tuple[int, float]]:
        """Retrieve authorship attribution of an article.

        This method uses WikiHistory to retrieve the authors measured by
        character count.

        Sample:

        >>> import pywikibot
        >>> site = pywikibot.Site('wikipedia:en')
        >>> page = pywikibot.Page(site, 'Pywikibot')
        >>> auth = page.authorship()  # doctest: +SKIP
        >>> auth  # doctest: +SKIP
        {'1234qwer1234qwer4': (68, 100.0)}

        .. important:: Only implemented for pages in Main, Project,
           Category and Template namespaces and only wikipedias of
           :attr:`WIKIBLAME_CODES` are supported.
        .. versionadded:: 9.3
           XTools is used to retrieve authors. This method replaces
           :meth:`main_authors`.
        .. versionchanged:: 10.1
           WikiHistory is used to retrieve authors due to :phab:`T392694`.

        Here are the differences between these two implementations:

        .. tabs::

           .. tab:: WikiHistory

              .. versionadded:: 10.1

              - Implemented from version 7.7 until 9.2 (with
                :meth:`main_authors` method) and from 10.1.
              - Main, Project, Category and Template namespaces are
                supported
              - Only 'als', 'bar', 'de', 'en', 'it', 'nds' and 'sco'
                Wikipedias are supported.
              - Revision ID *revid* or revision *date* is not supported.
                Always the latest revision is used.
              - Only the most 5 authors are given.
              - No additional parsing library is required.


              .. seealso::
                 - https://wikihistory.toolforge.org
                 - https://de.wikipedia.org/wiki/WP:HT/wikihistory

           .. tab:: XTools

              .. versionremoved:: 10.1

              - Implemented from version 9.3 until 10.0.
              - Only Main namespace is supported.
              - Only 'ar', 'de', 'en', 'es', 'eu', 'fr', 'hu', 'id',
                'it', 'ja', 'nl', 'pl', 'pt' and 'tr' Wikipedias are
                supported.
              - Revision ID *revid* or revision *date* is supported to
                get authorship for this revision.
              - All authors can be given.
              - wikitextparser parsing library is required.

              .. seealso::
                 - https://xtools.wmcloud.org/authorship/
                 - https://www.mediawiki.org/wiki/XTools/Authorship
                 - https://www.mediawiki.org/wiki/WikiWho


        :param n: Only return the first *n* or fewer authors.
        :param min_chars: Only return authors with more than *min_chars*
            chars changes.
        :param min_pct: Only return authors with more than *min_pct*
            percentage edits.
        :param max_pct_sum: Only return authors until the prcentage sum
            reached *max_pct_sum*.
        :return: Character count and percentage of edits for each
            username.

        :raise NotImplementedError: unsupported site or unsupported
            namespace.
        :raise NoPageError: The page does not exist.
        :raise TimeoutError: WikiHistory timeout
        """
        if n and n > 5:
            warn('Only the first 5 authors can be given.', stacklevel=2)

        baseurl = 'https://wikihistory.toolforge.org'
        pattern = (r'><bdi>(?P<author>.+?)</bdi></a>\s'
                   r'\((?P<percent>\d{1,3})&')

        self._check_wh_supported()

        for onlynew in (1, 0):
            url = baseurl + (f'/wiki/getauthors.php?wiki={self.site.code}wiki'
                             f'&page_id={self.pageid}&onlynew={onlynew}')

            r = pywikibot.comms.http.fetch(url)
            if r.status_code != HTTPStatus.OK:
                r.raise_for_status()

            if 'Timeout' not in r.text:
                break

            pywikibot.sleep(pywikibot.config.retry_wait)
        else:
            raise pywikibot.exceptions.TimeoutError('WikiHistory Timeout')

        length = len(self.text)
        result: list[list[str]] = []
        pct_sum = 0.0
        for rank, (user, cnt) in enumerate(re.findall(pattern, r.text),
                                           start=1):
            chars = length * int(cnt) // 100
            percent = float(cnt)

            # take into account that data() is ordered
            if n and rank > n or chars < min_chars or percent < min_pct:
                break

            result.append((user, chars, percent))

            pct_sum += percent
            if max_pct_sum and pct_sum >= max_pct_sum:
                break

        return {user: (chars, percent) for user, chars, percent in result}



"""Family module for Wiktionary."""
#
# (C) Pywikibot team, 2005-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


# The Wikimedia family that is known as Wiktionary
class Family(family.SubdomainFamily, family.WikimediaFamily):

    """Family class for Wiktionary.

    .. versionchanged:: 8.0
       ``alphabetic_sv`` attribute was removed; ``interwiki_putfirst``
       attribute was removed and default setting from parent class is
       used.
    """

    name = 'wiktionary'

    closed_wikis = [
        # https://noc.wikimedia.org/conf/highlight.php?file=dblists/closed.dblist
        'aa', 'ab', 'ak', 'as', 'av', 'bh', 'bi', 'bm', 'bo', 'ch', 'cr', 'dz',
        'ik', 'mh', 'pi', 'rm', 'rn', 'sc', 'sn', 'to', 'tw', 'xh', 'yo', 'za',
    ]

    removed_wikis = [
        # https://noc.wikimedia.org/conf/highlight.php?file=dblists/deleted.dblist
        'als', 'ba', 'dk', 'mo', 'tlh', 'tokipona',
    ]

    codes = {
        'af', 'am', 'an', 'ang', 'ar', 'ast', 'ay', 'az', 'bcl', 'be', 'bew',
        'bg', 'bjn', 'blk', 'bn', 'br', 'bs', 'btm', 'ca', 'chr', 'ckb', 'co',
        'cs', 'csb', 'cy', 'da', 'de', 'diq', 'dv', 'el', 'en', 'eo', 'es',
        'et', 'eu', 'fa', 'fi', 'fj', 'fo', 'fr', 'fy', 'ga', 'gd', 'gl',
        'gn', 'gom', 'gor', 'gu', 'guw', 'gv', 'ha', 'he', 'hi', 'hif', 'hr',
        'hsb', 'hu', 'hy', 'ia', 'id', 'ie', 'ig', 'io', 'is', 'it', 'iu',
        'ja', 'jbo', 'jv', 'ka', 'kaa', 'kbd', 'kcg', 'kk', 'kl', 'km', 'kn',
        'ko', 'ks', 'ku', 'kw', 'ky', 'la', 'lb', 'li', 'lmo', 'ln', 'lo',
        'lt', 'lv', 'mad', 'mg', 'mi', 'min', 'mk', 'ml', 'mn', 'mni', 'mnw',
        'mr', 'ms', 'mt', 'my', 'na', 'nah', 'nds', 'ne', 'nia', 'nl', 'nn',
        'no', 'oc', 'om', 'or', 'pa', 'pl', 'pnb', 'ps', 'pt', 'qu', 'ro',
        'roa-rup', 'ru', 'rw', 'sa', 'sat', 'scn', 'sd', 'sg', 'sh', 'shn',
        'shy', 'si', 'simple', 'sk', 'skr', 'sl', 'sm', 'so', 'sq', 'sr', 'ss',
        'st', 'su', 'sv', 'sw', 'ta', 'tcy', 'te', 'tg', 'th', 'ti', 'tk',
        'tl', 'tn', 'tpi', 'tr', 'ts', 'tt', 'ug', 'uk', 'ur', 'uz', 'vec',
        'vi', 'vo', 'wa', 'wo', 'yi', 'yue', 'zgh', 'zh', 'zh-min-nan', 'zu',
    }

    category_redirect_templates = {
        '_default': (),
        'ar': ('تحويل تصنيف',),
        'zh': ('分类重定向',),
    }

    # Global bot allowed languages on
    # https://meta.wikimedia.org/wiki/BPI#Current_implementation
    # & https://meta.wikimedia.org/wiki/Special:WikiSets/2
    cross_allowed = [
        'af', 'am', 'an', 'ang', 'ar', 'ast', 'ay', 'az', 'bcl', 'be', 'bg',
        'bjn', 'bn', 'br', 'bs', 'ca', 'chr', 'co', 'cs', 'csb', 'cy', 'da',
        'diq', 'dv', 'el', 'eo', 'es', 'et', 'eu', 'fa', 'fi', 'fj', 'fo',
        'ga', 'gd', 'gl', 'gn', 'gom', 'gor', 'gu', 'guv', 'gv', 'ha', 'hiv',
        'hr', 'hsb', 'hu', 'hy', 'ia', 'id', 'ie', 'ig', 'io', 'iu', 'jbo',
        'jv', 'ka', 'kk', 'kl', 'km', 'kn', 'ko', 'ks', 'ku', 'kw', 'ky', 'la',
        'lb', 'lmo', 'ln', 'lo', 'lt', 'lv', 'mg', 'mi', 'min', 'mk', 'mn',
        'mni', 'mnw', 'mr', 'ms', 'mt', 'my', 'na', 'nah', 'nds', 'ne', 'nia',
        'nl', 'nn', 'no', 'oc', 'om', 'or', 'pa', 'pnb', 'ps', 'pt', 'qu',
        'roa-rup', 'rw', 'sa', 'sat', 'scn', 'sd', 'sg', 'sh', 'shn', 'shy',
        'si', 'simple', 'sk', 'skr', 'sl', 'sm', 'so', 'sq', 'sr', 'ss', 'st',
        'su', 'sv', 'sw', 'ta', 'te', 'tg', 'th', 'ti', 'tk', 'tl', 'tn',
        'tpi', 'tr', 'ts', 'tt', 'ug', 'uk', 'ur', 'uz', 'vec', 'vi', 'vo',
        'wa', 'wo', 'yi', 'zh', 'zh-min-nan', 'zu',
    ]

    interwiki_on_one_line = ['pl']

    interwiki_attop = ['pl']

    # Subpages for documentation.
    # TODO: List is incomplete, to be completed for missing languages.
    doc_subpages = {
        '_default': (('/doc', ),
                     ['en']
                     ),
        'ar': ('/شرح', '/doc'),
        'sr': ('/док', ),
    }

    @classmethod
    def __post_init__(cls) -> None:
        """Add 'zh-yue' code alias due to :phab:`T341960`.

        .. versionadded:: 8.3
        """
        aliases = cls.code_aliases.copy()
        aliases['zh-yue'] = 'yue'
        cls.code_aliases = aliases



"""Base for terminal user interfaces."""
#
# (C) Pywikibot team, 2003-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import getpass
import logging
import re
import sys
import threading
from typing import Any, Literal, NoReturn, TextIO

import pywikibot
from pywikibot import config
from pywikibot.backports import (
    Iterable,
    RLock,
    Sequence,
    batched,
    removeprefix,
)
from pywikibot.bot_choice import (
    ChoiceException,
    Option,
    OutputOption,
    QuitKeyboardInterrupt,
    StandardOption,
)
from pywikibot.logging import INFO, INPUT, STDOUT, VERBOSE, WARNING
from pywikibot.userinterfaces import transliteration
from pywikibot.userinterfaces._interface_base import ABUIC


transliterator = transliteration.Transliterator(config.console_encoding)

#: Colors supported by Pywikibot
colors = [
    'default',
    'black',
    'blue',
    'green',
    'aqua',
    'red',
    'purple',
    'yellow',
    'lightgray',
    'gray',
    'lightblue',
    'lightgreen',
    'lightaqua',
    'lightred',
    'lightpurple',
    'lightyellow',
    'white',
]

colorTagR = re.compile(
    '<<((?:{0})(?:;(?:{0}))?)>>'.format('|'.join([*colors, 'previous'])))


class UI(ABUIC):

    """Base for terminal user interfaces.

    .. versionchanged:: 6.2:
       subclassed from
       :py:obj:`userinterfaces._interface_base.ABUIC`
    """

    split_col_pat = re.compile(r'(\w+);?(\w+)?')

    def __init__(self) -> None:
        """Initialize the UI.

        This caches the std-streams locally so any attempts to
        monkey-patch the streams later will not work.

        .. versionchanged:: 7.1
           memorize original streams
        """
        # for Windows GUI they can be None under some conditions
        self.stdin = sys.__stdin__ or sys.stdin
        self.stdout = sys.__stdout__ or sys.stdout
        self.stderr = sys.__stderr__ or sys.stderr

        self.argv = sys.argv
        self.encoding = config.console_encoding
        self.transliteration_target = config.transliteration_target
        self.cache = []
        self.lock = RLock()

    def init_handlers(
        self,
        root_logger,
        default_stream: TextIO | Literal['stderr', 'stdout'] = 'stderr'
    ) -> None:
        """Initialize the handlers for user output.

        This method initializes handler(s) for output levels VERBOSE (if
        enabled by config.verbose_output), INFO, STDOUT, WARNING, ERROR,
        and CRITICAL. STDOUT writes its output to sys.stdout; all the
        others write theirs to sys.stderr.
        """
        if default_stream == 'stdout':
            default_stream = self.stdout
        elif default_stream == 'stderr':
            default_stream = self.stderr

        # default handler for display to terminal
        default_handler = TerminalHandler(self, stream=default_stream)
        if config.verbose_output:
            default_handler.setLevel(VERBOSE)
        else:
            default_handler.setLevel(INFO)
        # this handler ignores levels above INPUT
        default_handler.addFilter(MaxLevelFilter(INPUT))
        default_handler.setFormatter(
            logging.Formatter(fmt='%(message)s%(newline)s'))
        root_logger.addHandler(default_handler)

        # handler for level STDOUT
        output_handler = TerminalHandler(self, stream=self.stdout)
        output_handler.setLevel(STDOUT)
        output_handler.addFilter(MaxLevelFilter(STDOUT))
        output_handler.setFormatter(
            logging.Formatter(fmt='%(message)s%(newline)s'))
        root_logger.addHandler(output_handler)

        # handler for levels WARNING and higher
        warning_handler = TerminalHandler(self, stream=self.stderr)
        warning_handler.setLevel(WARNING)
        warning_handler.setFormatter(
            logging.Formatter(fmt='%(levelname)s: %(message)s%(newline)s'))
        root_logger.addHandler(warning_handler)

        warnings_logger = logging.getLogger('py.warnings')
        warnings_logger.addHandler(warning_handler)

    def encounter_color(self, color, target_stream) -> NoReturn:
        """Abstract method to handle the next color encountered."""
        raise NotImplementedError(f'The {type(self).__name__} class does not'
                                  ' support colors.')

    @classmethod
    def divide_color(cls, color):
        """Split color label in a tuple.

        Received color is a string like 'fg_color;bg_color' or
        'fg_color'. Returned values are (fg_color, bg_color) or
        (fg_color, None).
        """
        return cls.split_col_pat.search(color).groups()

    def _write(self, text: str, target_stream) -> None:
        """Write the text to the target stream.

        sys.stderr and sys.stdout are frozen upon initialization to
        original streams (which are stored in the sys module as
        `sys.__stderr__` and `sys.__stdout__`). This works fine except
        when using `redirect_stderr` or `redirect_stdout` context
        managers, where values of global `sys.stderr` and `sys.stdout`
        are temporarily changed to a redirecting `StringIO` stream, in
        which case writing to the frozen streams (which are still set to
        `sys.__stderr__` / `sys.__stdout__`) will not write to the
        redirecting stream as expected. So, we check the target stream
        against the frozen streams, and then write to the (potentially
        redirected) `sys.stderr` or `sys.stdout` stream.

        .. versionchanged:: 7.1
           instead of writing to `target_stream`, dispatch to
           `sys.stderr` or `sys.stdout`.
        """
        if target_stream == self.stderr:
            sys.stderr.write(text)
        elif target_stream == self.stdout:
            sys.stdout.write(text)
        else:  # pragma: no cover
            try:
                out, err = self.stdout.name, self.stderr.name
            except AttributeError:
                out, err = self.stdout, self.stderr
            raise OSError(f'Target stream {target_stream.name} is neither '
                          f'stdin ({out}) nor stderr ({err})')

    def support_color(self, target_stream) -> bool:
        """Return whether the target stream does support colors."""
        return False

    def _print(self, text, target_stream) -> None:
        """Write the text to the target stream handling the colors."""
        colorized = (config.colorized_output
                     and self.support_color(target_stream))
        colored_line = False
        # Color tags might be cascaded, e.g. because of transliteration.
        # Therefore we need this stack.
        color_stack = ['default']
        text_parts = colorTagR.split(text)

        # Add default before the last linefeed
        if text.endswith('\n'):
            text_parts[-1] = re.sub(r'\r?\n\Z', '', text_parts[-1])
            text_parts.extend(('default', '\n'))

        text_parts.append('default')

        len_text_parts = len(text_parts) // 2
        for index, (txt, next_color) in enumerate(batched(text_parts, 2,
                                                          strict=True)):
            current_color = color_stack[-1]
            if next_color == 'previous':
                if len(color_stack) > 1:  # keep the last element in the stack
                    color_stack.pop()
                next_color = color_stack[-1]
            else:
                color_stack.append(next_color)

            if current_color != next_color:
                colored_line = True

            if colored_line and not colorized:
                if '\n' in txt:  # Normal end of line
                    txt = txt.replace('\n', ' ***\n', 1)
                    colored_line = False
                elif index == len_text_parts - 1:  # Or end of text
                    txt += ' ***'
                    colored_line = False

            # print the text up to the tag.
            self._write(txt, target_stream)

            if current_color != next_color and colorized:
                # set the new color, but only if they change
                # flush first to enable color change on Windows (T283808)
                target_stream.flush()
                self.encounter_color(color_stack[-1], target_stream)

        # finally flush stream to show the text, required for Windows (T303373)
        target_stream.flush()

    def output(self, text, targetStream=None) -> None:
        """Forward text to cache and flush if output is not locked.

        All input methods locks the output to a stream but collect them
        in cache. They will be printed with next unlocked output call or
        at termination time.

        .. versionchanged:: 7.0
           Forward text to cache and flush if output is not locked.
        """
        self.cache_output(text, targetStream=targetStream)
        if not self.lock.locked():
            self.flush()

    def flush(self) -> None:
        """Output cached text.

        .. versionadded:: 7.0
        """
        with self.lock:
            for args, kwargs in self.cache:
                self.stream_output(*args, **kwargs)
            self.cache.clear()

    def cache_output(self, *args, **kwargs) -> None:
        """Put text to cache.

        .. versionadded:: 7.0
        """
        with self.lock:
            self.cache.append((args, kwargs))

    def stream_output(self, text, targetStream=None) -> None:
        """Output text to a stream.

        If a character can't be displayed in the encoding used by the user's
        terminal, it will be replaced with a question mark or by a
        transliteration.

        .. versionadded:: 7.0
           ``UI.output()`` was renamed to ``UI.stream_output()``
        """
        if config.transliterate:
            # Encode our unicode string in the encoding used by the user's
            # console, and decode it back to unicode. Then we can see which
            # characters can't be represented in the console encoding.
            # We need to take min(console_encoding, transliteration_target)
            # the first is what the terminal is capable of
            # the second is how unicode-y the user would like the output
            codecedText = text.encode(self.encoding,
                                      'replace').decode(self.encoding)
            if self.transliteration_target:
                codecedText = codecedText.encode(
                    self.transliteration_target,
                    'replace').decode(self.transliteration_target)
            transliteratedText = ''
            # Note: A transliteration replacement might be longer than the
            # original character, e.g. ч is transliterated to ch.
            prev = '-'
            for i, char in enumerate(codecedText):
                # work on characters that couldn't be encoded, but not on
                # original question marks.
                if char == '?' and text[i] != '?':
                    try:
                        transliterated = transliterator.transliterate(
                            text[i], default='?', prev=prev, succ=text[i + 1])
                    except IndexError:
                        transliterated = transliterator.transliterate(
                            text[i], default='?', prev=prev, succ=' ')
                    # transliteration was successful. The replacement
                    # could consist of multiple letters.
                    # mark the transliterated letters in yellow.
                    transliteratedText = (
                        f'{transliteratedText}'
                        f'<<lightyellow>>{transliterated}<<previous>>'
                    )
                    # memorize if we replaced a single letter by multiple
                    # letters.
                    if transliterated:
                        prev = transliterated[-1]
                else:
                    # no need to try to transliterate.
                    transliteratedText += char
                    prev = char
            text = transliteratedText

        if not targetStream:
            targetStream = self.stderr

        self._print(text, targetStream)

    def _raw_input(self):
        # May be overridden by subclass
        return input()

    def input(self, question: str,
              password: bool = False,
              default: str | None = '',
              force: bool = False) -> str:
        """Ask the user a question and return the answer.

        Works like raw_input(), but returns a unicode string instead of
        ASCII.

        Unlike raw_input, this function automatically adds a colon and
        space after the question if they are not already present. Also
        recognises a trailing question mark.

        :param question: The question, without trailing whitespace.
        :param password: if True, hides the user's input (for password
            entry).
        :param default: The default answer if none was entered. None to
            require an answer.
        :param force: Automatically use the default
        """
        assert not password or not default

        question = question.strip()
        end_marker = question[-1]
        if end_marker in (':', '?'):
            question = question[:-1]
        else:
            end_marker = ':'

        if default:
            question += f' (default: {default})'
        question += end_marker

        # lock stream output
        with self.lock:
            if force:
                self.stream_output(question + '\n')
                return default

            # sound the terminal bell to notify the user
            if config.ring_bell:
                sys.stdout.write('\07')

            # TODO: make sure this is logged as well
            while True:
                self.stream_output(question + ' ')
                text = self._input_reraise_cntl_c(password)

                if text is None:
                    continue

                if text:
                    return text

                if default is not None:
                    return default

    def _input_reraise_cntl_c(self, password):
        """Input and decode, and re-raise Control-C."""
        try:
            if password:
                # Python 3 requires that stderr gets flushed, otherwise is the
                # message only visible after the query.
                self.stderr.flush()
                text = getpass.getpass('')
            else:
                text = self._raw_input()
        except KeyboardInterrupt:
            raise QuitKeyboardInterrupt
        except UnicodeDecodeError:
            return None  # wrong terminal encoding, T258143
        return text

    def input_choice(
        self,
        question: str,
        options: Iterable[tuple[str, str] | Option] | Option,
        default: str | None = None,
        return_shortcut: bool = True,
        automatic_quit: bool = True,
        force: bool = False,
    ) -> Any:
        """Ask the user and returns a value from the options.

        Depending on the options setting *return_shortcut* to False may
        not be sensible when the option supports multiple values as
        it'll return an ambiguous index.

        .. versionchanged:: 9.0
           Raise ValueError if no *default* value is given with *force*;
           raise ValueError if *force* is True and *default* value is
           invalid; raise TypeError if *default* value is neither str
           nor None.

        :param question: The question, without trailing whitespace.
        :param options: Iterable of all available options. Each entry contains
            the full length answer and a shortcut of only one character.
            Alternatively they may be Option (or subclass) instances or
            ChoiceException instances which have a full option and shortcut
            and will be raised if selected.
        :param default: The default answer if no was entered. None to require
            an answer.
        :param return_shortcut: Whether the shortcut or the index in the option
            should be returned.
        :param automatic_quit: Adds the option 'Quit' ('q') if True and throws
            a :py:obj:`QuitKeyboardInterrupt` if selected.
        :param force: Automatically use the default
        :return: If return_shortcut the shortcut of options or the value of
            default (if it's not None). Otherwise the index of the answer in
            options. If default is not a shortcut, it'll return -1.
        :raises ValueError: invalid or no *default* value is given with
            *force* or no or an invalid option is given.
        :raises TypeError: *default* value is neither None nor str
        """
        def output_option(option, before_question) -> None:
            """Print an OutputOption before or after question."""
            if isinstance(option, OutputOption) \
               and option.before_question is before_question:
                self.stream_output(option.out + '\n')

        if force and not default:
            raise ValueError('With no default option it cannot be forced')

        # make a copy
        options = [options] if isinstance(options, Option) else list(options)

        if not options:
            raise ValueError('No options are given.')
        if automatic_quit:
            options.append(QuitKeyboardInterrupt())

        if isinstance(default, str):
            default = default.lower()
        elif default is not None:
            raise TypeError(f'Invalid type {type(default).__name__!r} for '
                            f"parameter 'default' ({default}); str expected")

        for i, option in enumerate(options):
            if not isinstance(option, Option):
                if len(option) != 2:
                    raise ValueError(f'Option #{i} does not consist of an '
                                     'option and shortcut.')
                options[i] = StandardOption(*option)
            # TODO: Test for uniquity

        handled = False

        # lock stream output
        with self.lock:
            while not handled:
                for option in options:
                    output_option(option, before_question=True)
                output = Option.formatted(question, options, default)
                if force:
                    self.stream_output(output + '\n')
                    answer = default
                else:
                    answer = self.input(output) or default
                # something entered or default is defined
                if answer:
                    for index, option in enumerate(options):
                        if option.handled(answer):
                            answer = option.result(answer)
                            output_option(option, before_question=False)
                            handled = option.stop
                            break
                    else:
                        if force:
                            raise ValueError(
                                f'{default!r} is not a valid Option for '
                                f'{removeprefix(output, question).lstrip()}')

        if isinstance(answer, ChoiceException):
            raise answer
        if not return_shortcut:
            return index
        return answer

    def input_list_choice(self, question: str, answers: Sequence[Any],
                          default: int | str | None = None,
                          force: bool = False) -> Any:
        """Ask the user to select one entry from a list of entries.

        :param question: The question, without trailing whitespace.
        :param answers: A sequence of options to be chosen.
        :param default: The default answer if no was entered. None to
            require an answer.
        :param force: Automatically use the default.
        :return: Return a single Sequence entry.
        """
        # lock stream output
        with self.lock:
            if not force:
                line_template = f'{{0: >{len(str(len(answers)))}}}: {{1}}\n'
                for i, entry in enumerate(answers, start=1):
                    self.stream_output(line_template.format(i, entry))

            while True:
                choice = self.input(question, default=default, force=force)

                try:
                    parsedchoice = int(choice) - 1
                except (TypeError, ValueError):
                    if choice in answers:
                        return choice
                    parsedchoice = -1

                # User typed choice number
                if 0 <= parsedchoice < len(answers):
                    return answers[parsedchoice]

                if force:
                    raise ValueError(
                        f'Invalid value "{default}" for default during force.')

                self.stream_output('Error: Invalid response\n')

    @staticmethod
    def editText(text: str,
                 jumpIndex: int | None = None,
                 highlight: str | None = None) -> str | None:
        """Return the text as edited by the user.

        Uses a Tkinter edit box because we don't have a console editor

        :param text: the text to be edited
        :param jumpIndex: position at which to put the caret
        :param highlight: each occurrence of this substring will be
            highlighted
        :return: the modified text, or None if the user didn't save the
            text file in his text editor
        """
        try:
            from pywikibot.userinterfaces import gui
        except ImportError as e:
            pywikibot.warning(f'Could not load GUI modules: {e}')
            return text
        editor = gui.EditBoxWindow()
        return editor.edit(text, jumpIndex=jumpIndex, highlight=highlight)


class TerminalHandler(logging.StreamHandler):

    """A handler class that writes logging records to a terminal.

    This class does not close the stream, as sys.stdout or sys.stderr
    may be (and usually will be) used.

    Slightly modified version of the StreamHandler class that ships with
    logging module, plus code for colorization of output.
    """

    # create a class-level lock that can be shared by all instances
    sharedlock = threading.RLock()

    def __init__(self, UI, stream=None) -> None:
        """Initialize the handler.

        If stream is not specified, sys.stderr is used.
        """
        super().__init__(stream=stream)
        self.UI = UI

    def createLock(self) -> None:
        """Acquire a thread lock for serializing access to the underlying I/O.

        Replace Handler's instance-specific lock with the shared class
        lock to ensure that only one instance of this handler can write
        to the console at a time.
        """
        self.lock = TerminalHandler.sharedlock

    def emit(self, record) -> None:
        """Emit the record formatted to the output."""
        self.flush()
        if record.name == 'py.warnings':
            # Each warning appears twice
            # the second time it has a 'message'
            if 'message' in record.__dict__:
                return

            record.__dict__.setdefault('newline', '\n')

        msg = self.format(record)
        self.UI.output(msg, targetStream=self.stream)


class MaxLevelFilter:

    """Filter that only passes records at or below a specific level.

    .. note:: setting handler level only passes records at or *above* a
       specified level, so this provides the opposite functionality.
    """

    def __init__(self, level=None) -> None:
        """Initializer."""
        self.level = level

    def filter(self, record):
        """Return true if the level is below or equal to the set level."""
        if self.level:
            return record.levelno <= self.level
        return True



"""Objects representing interwiki map of MediaWiki site."""
#
# (C) Pywikibot team, 2015-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import pywikibot


class _IWEntry:

    """An entry of the _InterwikiMap with a lazy loading site."""

    def __init__(self, local, url, prefix=None) -> None:
        self._site = None
        self.local = local
        self.url = url
        self.prefix = prefix

    @property
    def site(self):
        if self._site is None:
            try:
                self._site = pywikibot.Site(
                    url=self.url, fam=None if self.local else self.prefix)
            except Exception as e:
                self._site = e
        return self._site


class _InterwikiMap:

    """A representation of the interwiki map of a site."""

    def __init__(self, site) -> None:
        """Create an empty uninitialized interwiki map for the given site.

        :param site: Given site for which interwiki map is to be created
        :type site: pywikibot.site.APISite
        """
        super().__init__()
        self._site = site
        self._map = None

    def reset(self) -> None:
        """Remove all mappings to force building a new mapping."""
        self._map = None

    @property
    def _iw_sites(self):
        """Fill the interwikimap cache with the basic entries."""
        # _iw_sites is a local cache to return an APISite instance depending
        # on the interwiki prefix of that site
        if self._map is None:
            self._map = {iw['prefix']: _IWEntry('local' in iw,
                                                iw['url'],
                                                iw['prefix'])
                         for iw in self._site.siteinfo['interwikimap']}
        return self._map

    def __getitem__(self, prefix):
        """Return the site, locality and url for the requested prefix.

        :param prefix: Interwiki prefix
        :type prefix: Dictionary key
        :rtype: _IWEntry
        :raises KeyError: Prefix is not a key
        :raises TypeError: Site for the prefix is of wrong type
        """
        if prefix not in self._iw_sites:
            raise KeyError(f"'{prefix}' is not an interwiki prefix.")
        if isinstance(self._iw_sites[prefix].site, pywikibot.site.BaseSite):
            return self._iw_sites[prefix]
        if isinstance(self._iw_sites[prefix].site, Exception):
            raise self._iw_sites[prefix].site
        raise TypeError(f'_iw_sites[{prefix}] is wrong type: '
                        f'{type(self._iw_sites[prefix].site)}')

    def get_by_url(self, url: str) -> set[str]:
        """Return a set of prefixes applying to the URL.

        :param url: URL for the interwiki
        """
        return {prefix for prefix, iw_entry in self._iw_sites.items()
                if iw_entry.url == url}



"""User interface for Unix terminals."""
#
# (C) Pywikibot team, 2003-2022
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import re

from pywikibot.userinterfaces import terminal_interface_base


unixColors = {
    'default':     chr(27) + '[0m',   # Unix end tag to switch back to default
    'black':       chr(27) + '[30m',  # Black start tag
    'red':         chr(27) + '[31m',  # Red start tag
    'green':       chr(27) + '[32m',  # Green start tag
    'yellow':      chr(27) + '[33m',  # Yellow start tag
    'blue':        chr(27) + '[34m',  # Blue start tag
    'purple':      chr(27) + '[35m',  # Purple start tag (Magenta)
    'aqua':        chr(27) + '[36m',  # Aqua start tag (Cyan)
    'lightgray':   chr(27) + '[37m',  # Light gray start tag (White)
    'gray':        chr(27) + '[90m',  # Gray start tag
    'lightred':    chr(27) + '[91m',  # Light Red tag
    'lightgreen':  chr(27) + '[92m',  # Light Green tag
    'lightyellow': chr(27) + '[93m',  # Light Yellow tag
    'lightblue':   chr(27) + '[94m',  # Light Blue tag
    'lightpurple': chr(27) + '[95m',  # Light Purple tag (Magenta)
    'lightaqua':   chr(27) + '[96m',  # Light Aqua tag (Cyan)
    'white':       chr(27) + '[97m',  # White start tag (Bright White)
}


class UnixUI(terminal_interface_base.UI):

    """User interface for Unix terminals."""

    def support_color(self, target_stream) -> bool:
        """Return that the target stream supports colors."""
        return True

    @staticmethod
    def make_unix_bg_color(color):
        """Obtain background color from foreground color."""
        code = re.search(r'(?<=\[)\d+', color).group()
        return chr(27) + '[' + str(int(code) + 10) + 'm'

    def encounter_color(self, color, target_stream) -> None:
        """Write the Unix color directly to the stream."""
        fg, bg = self.divide_color(color)
        fg = unixColors[fg]
        self._write(fg, target_stream)
        if bg is not None:
            bg = unixColors[bg]
            self._write(self.make_unix_bg_color(bg), target_stream)



"""Family module for Lingua Libre.

.. versionadded:: 6.5
"""
#
# (C) Pywikibot team, 2021-2024
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

from pywikibot import family


class Family(family.WikimediaFamily, family.WikibaseFamily):

    """Family class for Lingua Libre.

    .. versionadded:: 6.5
    """

    name = 'lingualibre'

    langs = {
        'lingualibre': 'lingualibre.org'
    }

    interwiki_forward = 'wikipedia'

    def scriptpath(self, code) -> str:
        """Return the script path for this family."""
        return ''



#!/usr/bin/env python3
"""Script to log the bot in to a wiki account.

Suggestion is to make a special account to use for bot use only. Make
sure this bot account is well known on your home wiki before using.

The following parameters are supported:

-all         Try to log in on all sites where a username is defined in
             user config file (user-config.py).

-cookies     Only login from cookies file, don't ask for password

-logout      Log out of the current site. Combine with ``-all`` to log
             out of all sites, or with :ref:`global options` ``-family``,
             ``-lang`` or ``-site`` to log out of a specific site.

-oauth       Generate OAuth authentication information.

             .. note:: Need to copy OAuth tokens to your user config
                file manually. -logout is not compatible with -oauth.

-autocreate  Auto-create an account using unified login when necessary.

             .. note:: the global account must exist already before
                using this.

-async       Run the bot in parallel tasks, only useful together with
             ``-all`` option

.. hint:: Use :ref:`global options` ``-code``, ``-family`` or ``-site``
   to determine the site to login/logout.

If not given as parameter, the script will ask for your username and
password (password entry will be hidden), log in to your home wiki using
this combination, and store the resulting cookies (containing your
password hash, so keep it secured!) in a file in the data subdirectory.

All scripts in this library will be looking for this cookie file and
will use the login information if it is present.

To log out, throw away the ``*.lwp`` file that is created in the data
subdirectory.

.. versionchanged:: 7.4
   moved to :mod:`pywikibot.scripts` folder.
.. versionchanged:: 7.7
   *-async* option was added.
.. versionchanged:: 10.2
   wildcard site codes in ``usernames`` dict are supported.
.. versionchanged:: 10.3
   the -cookies option
"""
#
# (C) Pywikibot team, 2003-2025
#
# Distributed under the terms of the MIT license.
#
from __future__ import annotations

import datetime
from contextlib import nullcontext, suppress

import pywikibot
from pywikibot import config
from pywikibot.exceptions import NoUsernameError, SiteDefinitionError
from pywikibot.family import Family
from pywikibot.login import OauthLoginManager
from pywikibot.tools.threading import BoundedPoolExecutor


def _get_consumer_token(site) -> tuple[str, str]:
    key_msg = f'OAuth consumer key on {site.code}:{site.family}'
    key = pywikibot.input(key_msg)
    secret_msg = f'OAuth consumer secret for consumer {key}'
    secret = pywikibot.input(secret_msg, password=True)
    return key, secret


def _oauth_login(site) -> None:
    consumer_key, consumer_secret = _get_consumer_token(site)
    login_manager = OauthLoginManager(consumer_secret, site, consumer_key)
    login_manager.login()
    identity = login_manager.identity
    if identity is None:
        pywikibot.error(f'Invalid OAuth info for {site}.')
    elif site.username() != identity['username']:
        pywikibot.error(
            'Logged in on {site} via OAuth as {wrong}, but expect as {right}'
            .format(site=site,
                    wrong=identity['username'], right=site.username()))
    else:
        oauth_token = login_manager.consumer_token + login_manager.access_token
        pywikibot.info(
            f'Logged in on {site} as {site.username()} via OAuth consumer '
            f'{consumer_key}\nNOTE: To use OAuth, you need to copy the'
            ' following line to your user config file:\n'
            f'authenticate[{site.hostname()!r}] = {oauth_token}'
        )


def login_one_site(code: str, family: str, oauth: bool, logout: bool,
                   autocreate: bool, cookies: bool) -> None:
    """Login on one site."""
    try:
        site = pywikibot.Site(code, family)
    except SiteDefinitionError:
        pywikibot.error(f'{family}:{code} is not a valid site, '
                        'please remove it from your user-config')
        return

    if oauth:
        _oauth_login(site)
        return

    if logout:
        site.logout()
    else:
        try:
            site.login(autocreate=autocreate, cookie_only=cookies)
        except NoUsernameError as e:
            pywikibot.error(e)

    user = site.user()
    if user:
        pywikibot.info(f'Logged in on {site} as {user}.')
    elif logout:
        pywikibot.info(f'Logged out of {site}.')
    else:
        pywikibot.info(f'Not logged in on {site}.')


def main(*args: str) -> None:
    """Process command line arguments and invoke bot.

    If args is an empty list, sys.argv is used.

    :param args: command line arguments
    """
    logall = False
    logout = False
    oauth = False
    autocreate = False
    asynchronous = False
    cookies = False
    unknown_args = []
    for arg in pywikibot.handle_args(args):
        if arg == '-all':
            logall = True
        elif arg == '-logout':
            logout = True
        elif arg == '-oauth':
            oauth = True
        elif arg == '-autocreate':
            autocreate = True
        elif arg == '-async':
            asynchronous = True
        elif arg == '-cookies':
            cookies = True
        else:
            unknown_args.append(arg)

    if pywikibot.bot.suggest_help(unknown_parameters=unknown_args):
        return

    if logall:
        namedict = config.usernames
    else:
        site = pywikibot.Site()
        namedict = {site.family.name: {site.code: None}}

    params = oauth, logout, autocreate, cookies
    context = (nullcontext(),
               BoundedPoolExecutor('ThreadPoolExecutor'))[asynchronous]
    with context as executor:
        for family_name in namedict:
            family_codes = namedict[family_name]

            if '*' in family_codes:
                user = family_codes['*']
                del family_codes['*']
                codes = (code for code in Family.load(family_name).codes
                         if code not in family_codes)
                family_codes.update(dict.fromkeys(codes, user))

            for code in family_codes:
                if asynchronous:
                    executor.submit(login_one_site, code, family_name, *params)
                else:
                    login_one_site(code, family_name, *params)


if __name__ == '__main__':
    start = datetime.datetime.now()
    with suppress(KeyboardInterrupt):
        main()
    pywikibot.info('\nExecution time: '
                   f'{(datetime.datetime.now() - start).seconds} seconds')
